# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-10-12 19:14
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0012_auto_20161012_1218'),
    ]

    operations = [
        migrations.AlterField(
            model_name='organization',
            name='source_url',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-10-14 20:48
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0022_auto_20161014_1350'),
    ]

    operations = [
        migrations.AlterField(
            model_name='actionrelatedentity',
            name='organization_ocd_id',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
        migrations.AlterField(
            model_name='actionrelatedentity',
            name='person_ocd_id',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
    ]



from django.conf.urls import url
from . import views
from . import feeds
from django.views.decorators.cache import never_cache # use never_cache wrapper for class-based Views which need to reflect current subscription status

from django.contrib import admin
admin.autodiscover() # XXX necessary?

urlpatterns = [
    url(r'^$', views.IndexView.as_view(), name='index'),
    url(r'^about/$', views.AboutView.as_view(), name='about'),
    url(r'^committees/$', views.CommitteesView.as_view(), name='committees'),
    url(r'^council-members/$', views.CouncilMembersView.as_view(),
        name='council_members'),

    url(r'^committee/(?P<slug>[^/]+)/$',
        never_cache(views.CommitteeDetailView.as_view()), name='committee_detail'),
    url(r'^committee/(?P<slug>[^/]+)/events/rss/$',
        feeds.CommitteeDetailEventsFeed(), name='committee_detail_events_feed'),
    url(r'^committee/(?P<slug>[^/]+)/actions/rss/$',
        feeds.CommitteeDetailActionFeed(), name='committee_detail_action_feed'),
    url(r'^committee/(?P<slug>[^/]+)/widget/$',
        views.CommitteeWidgetView.as_view(), name='committee_widget'),

    url(r'^legislation/(?P<slug>[^/]+)/$',
        never_cache(views.BillDetailView.as_view()), name='bill_detail'),
    url(r'^legislation/(?P<slug>[^/]+)/rss/$',
        feeds.BillDetailActionFeed(), name='bill_detail_action_feed'),
    url(r'^legislation/(?P<slug>[^/]+)/widget/$',
        views.BillWidgetView.as_view(), name='bill_widget'),

    url(r'^person/(?P<slug>[^/]+)/$', never_cache(views.PersonDetailView.as_view()), name='person'),
    url(r'^person/(?P<slug>[^/]+)/rss/$', feeds.PersonDetailFeed(), name='person_feed'),
    url(r'^person/(?P<slug>[^/]+)/widget/$',
        views.PersonWidgetView.as_view(), name='person_widget'),

    url(r'^events/$', never_cache(views.EventsView.as_view()), name='events'),
    url(r'^events/rss/$', feeds.EventsFeed(), name='events_feed'),
    url(r'^event/(?P<slug>.+)/$',
        views.EventDetailView.as_view(), name='event_detail'),

    url(r'^flush-cache/(.*)/$', views.flush, name='flush'),
    url(r'^pdfviewer/$', views.pdfviewer, name='pdfviewer'),
]




# -*- coding=utf-8 -*-

import json
import re
import datetime
import os
import sys

import requests
import pytz
import psycopg2
import sqlalchemy as sa

from requests.packages.urllib3.exceptions import InsecureRequestWarning

from dateutil import parser as date_parser

from django.core.management.base import BaseCommand
from django.core.exceptions import ImproperlyConfigured
from django.conf import settings
from django.utils.dateparse import parse_datetime, parse_date
from django.utils.text import slugify, Truncator
from django.db.utils import IntegrityError, DataError
from django.db.models import Max

from councilmatic_core.models import Person, Bill, Organization, Action, ActionRelatedEntity, \
    Post, Membership, Sponsorship, LegislativeSession, \
    Document, BillDocument, Event, EventParticipant, EventDocument, \
    EventAgendaItem

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

for configuration in ['OCD_JURISDICTION_ID',
                      'HEADSHOT_PATH',
                      'LEGISLATIVE_SESSIONS'
                      ]:

    if not hasattr(settings, configuration):
        raise ImproperlyConfigured(
            'You must define {0} in settings.py'.format(configuration))

if not (hasattr(settings, 'OCD_CITY_COUNCIL_ID') or hasattr(settings, 'OCD_CITY_COUNCIL_NAME')):
    raise ImproperlyConfigured(
        'You must define a OCD_CITY_COUNCIL_ID or OCD_CITY_COUNCIL_NAME in settings.py')

app_timezone = pytz.timezone(settings.TIME_ZONE)

DB_CONN = 'postgresql://{USER}:{PASSWORD}@{HOST}:{PORT}/{NAME}'

engine = sa.create_engine(DB_CONN.format(**settings.DATABASES['default']),
                          convert_unicode=True,
                          server_side_cursors=True)

if hasattr(settings, 'OCDAPI_BASE_URL'):
    base_url = settings.OCDAPI_BASE_URL
else:
    base_url = 'http://ocd.datamade.us'

if hasattr(settings, 'BOUNDARY_API_BASE_URL'):
    bndry_base_url = settings.BOUNDARY_API_BASE_URL
else:
    bndry_base_url = base_url

DEBUG = settings.DEBUG


class Command(BaseCommand):
    help = 'loads in data from the open civic data API'
    update_since = None

    def add_arguments(self, parser):
        parser.add_argument(
            '--endpoints',
            help="a specific endpoint to load data from",
            default='organizations,people,bills,events')

        parser.add_argument('--delete',
                            action='store_true',
                            default=False,
                            help='deletes all data, and then loads all legislative sessions (by default, this task does not delete data & only loads new/updated data from current legislative session)')

        parser.add_argument('--update_since',
                            help='Only update objects in the database that have changed since this date')

        parser.add_argument('--import_only',
                            action='store_true',
                            default=False,
                            help='Load already downloaded OCD data')

        parser.add_argument('--download_only',
                            action='store_true',
                            default=False,
                            help='Only download OCD data')

    def handle(self, *args, **options):

        self.connection = engine.connect()

        self.downloads_folder = 'downloads'
        self.this_folder = os.path.abspath(os.path.dirname(__file__))

        self.organizations_folder = os.path.join(self.downloads_folder, 'organizations')
        self.posts_folder = os.path.join(self.downloads_folder, 'posts')
        self.bills_folder = os.path.join(self.downloads_folder, 'bills')
        self.people_folder = os.path.join(self.downloads_folder, 'people')
        self.events_folder = os.path.join(self.downloads_folder, 'events')

        if options['update_since']:
            self.update_since = date_parser.parse(options['update_since'])

        endpoints = options['endpoints'].split(',')

        for endpoint in endpoints:

            if endpoint not in ['organizations', 'people', 'bills', 'events']:

                self.log_message('"{}" is not a valid endpoint'.format(endpoint), style='ERROR')

            else:

                download_only = options['download_only']
                import_only = options['import_only']

                if not import_only and not download_only:
                    download_only = True
                    import_only = True

                etl_method = getattr(self, '{}_etl'.format(endpoint))
                etl_method(import_only=import_only,
                           download_only=download_only,
                           delete=options['delete'])

        if getattr(settings, 'USING_NOTIFICATIONS', None):
            from django.core import management

            management.call_command('update_index', age=24)
            management.call_command('send_notifications')


    def log_message(self,
                    message,
                    fancy=False,
                    style='HTTP_SUCCESS',
                    art_file=None,
                    center=False,
                    timestamp=True):

        if timestamp:
            now = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            message = '{0} {1}'.format(now, message)

        if len(message) < 70 and center:
            padding = (70 - len(message)) / 2
            message = '{0}{1}{0}'.format(' ' * int(padding), message)

        if fancy and not art_file:
            thing_count = len(message) + 2

            message = '\n{0}\n  {1}  \n{0}'.format('-' * 70, message)

        elif art_file:
            art = open(os.path.join(self.this_folder, 'art', art_file)).read()
            message = '\n{0} \n {1}'.format(art, message)

        style = getattr(self.style, style)
        self.stdout.write(style('{}\n'.format(message)))

    def organizations_etl(self,
                          import_only=True,
                          download_only=True,
                          delete=False):

        if download_only:
            self.log_message('Downloading organizations ...',
                             center=True,
                             art_file='organizations.txt')

            self.grab_organizations()

        if import_only:
            self.log_message('Importing organizations ...',
                             center=True,
                             art_file='organizations.txt')

            self.insert_raw_organizations(delete=delete)
            self.insert_raw_posts(delete=delete)

            self.update_existing_organizations()
            self.update_existing_posts()

            self.add_new_organizations()
            self.add_new_posts()

        self.log_message('Organizations Complete!',
                         fancy=True,
                         center=True,
                         style='SUCCESS')

    def people_etl(self,
                   import_only=False,
                   download_only=False,
                   delete=False):

        if download_only:
            self.log_message('Downloading people ...',
                             center=True,
                             art_file='people.txt')

            self.grab_people()

        if import_only:

            self.log_message('Importing people ...',
                             center=True,
                             art_file='people.txt')

            self.insert_raw_people(delete=delete)
            self.insert_raw_memberships(delete=delete)

            self.update_existing_people()
            self.update_existing_memberships()

            self.add_new_people()
            self.add_new_memberships()

        self.log_message('People Complete!',
                         fancy=True,
                         center=True,
                         style='SUCCESS')

    def bills_etl(self,
                  import_only=False,
                  download_only=False,
                  delete=False):

        self.create_legislative_sessions()

        if download_only:
            self.log_message('Downloading bills ...',
                             center=True,
                             art_file='bills.txt')
            self.grab_bills()

        if import_only:
            self.log_message('Importing bills ...',
                             center=True,
                             art_file='bills.txt')
            self.insert_raw_bills(delete=delete)
            self.insert_raw_actions(delete=delete)

            self.update_existing_bills()
            self.update_existing_actions()

            self.add_new_bills()
            self.add_new_actions()

            self.insert_raw_action_related_entity(delete=delete)
            self.insert_raw_sponsorships(delete=delete)
            self.insert_raw_billdocuments(delete=delete)

            self.update_existing_action_related_entity()
            self.update_existing_sponsorships()
            self.update_existing_billdocuments()

            self.add_new_action_related_entity()
            self.add_new_sponsorships()
            self.add_new_billdocuments()

        self.log_message('Bills Complete!', fancy=True, style='SUCCESS', center=True)

    def events_etl(self,
                   import_only=False,
                   download_only=False,
                   delete=False):

        if download_only:
            self.log_message('Downloading events ...',
                             center=True,
                             fancy=True)
            self.grab_events()

        if import_only:
            self.log_message('Importing events ...',
                             center=True,
                             fancy=True)
            self.insert_raw_events(delete=delete)
            self.insert_raw_eventparticipants(delete=delete)
            self.insert_raw_eventdocuments(delete=delete)
            self.insert_raw_event_agenda_items(delete=delete)

            self.update_existing_events()
            self.update_existing_eventparticipants()
            self.update_existing_eventdocuments()
            self.update_existing_event_agenda_items()

            self.add_new_events()
            self.add_new_eventparticipants()
            self.add_new_eventdocuments()
            self.add_new_event_agenda_items()

        self.log_message('Events Complete!', fancy=True, style='SUCCESS', center=True)

    #########################
    ###                   ###
    ### DOWNLOAD FROM OCD ###
    ###                   ###
    #########################

    def grab_organizations(self):

        os.makedirs(self.organizations_folder, exist_ok=True)
        os.makedirs(self.posts_folder, exist_ok=True)

        # first grab city council root
        if hasattr(settings, 'OCD_CITY_COUNCIL_ID'):
            self.grab_organization_posts({'id': settings.OCD_CITY_COUNCIL_ID})
        else:
            self.grab_organization_posts({'name': settings.OCD_CITY_COUNCIL_NAME})

        # this grabs a paginated listing of all organizations within a
        # jurisdiction
        orgs_url = base_url + '/organizations/?jurisdiction_id=' + \
            settings.OCD_JURISDICTION_ID
        r = requests.get(orgs_url)
        page_json = json.loads(r.text)

        org_counter = 0
        post_counter = 0

        for i in range(page_json['meta']['max_page']):

            r = requests.get(orgs_url + '&page=' + str(i + 1))
            page_json = json.loads(r.text)

            org_counter += len(page_json['results'])

            for result in page_json['results']:

                post_count = self.grab_organization_posts({'id': result['id']})

                post_counter += post_count

                print('.', end='')
                sys.stdout.flush()

        print('\n')
        self.log_message('Downloaded {0} orgs and {1} posts'.format(org_counter, post_counter))

        # update relevant posts with shapes
        if hasattr(settings, 'BOUNDARY_SET') and settings.BOUNDARY_SET:
            self.populate_council_district_shapes()

    def grab_organization_posts(self, org_dict):
        url = base_url + '/organizations/'

        r = requests.get(url, params=org_dict)
        page_json = json.loads(r.text)
        organization_ocd_id = page_json['results'][0]['id']

        url = base_url + '/' + organization_ocd_id + '/'
        r = requests.get(url)
        page_json = json.loads(r.text)

        if page_json.get('error'):
            raise DataError(page_json['error'])

        ocd_uuid = org_dict['id'].split('/')[-1]
        organization_filename = '{}.json'.format(ocd_uuid)

        with open(os.path.join(self.organizations_folder, organization_filename), 'w') as f:
            f.write(json.dumps(page_json))


        for post_json in page_json['posts']:

            post_uuid = post_json['id'].split('/')[-1]
            post_filename = '{}.json'.format(post_uuid)
            post_json['org_ocd_id'] = org_dict['id']

            with open(os.path.join(self.posts_folder, post_filename), 'w') as f:
                f.write(json.dumps(post_json))

        for child in page_json['children']:
            self.grab_organization_posts({'id': child['id']})

        return len(page_json['posts'] + page_json['children'])

    def grab_people(self):
        # find people associated with existing organizations & bills

        os.makedirs(self.people_folder, exist_ok=True)


        counter = 0
        for organization_json in os.listdir(self.organizations_folder):

            org_info = json.load(open(os.path.join(self.organizations_folder, organization_json)))

            for membership_json in org_info['memberships']:
                person_json = self.grab_person_memberships(membership_json['person']['id'])

                person_uuid = person_json['id'].split('/')[-1]
                person_filename = '{}.json'.format(person_uuid)

                with open(os.path.join(self.people_folder, person_filename), 'w') as f:
                    f.write(json.dumps(person_json))

                print('.', end='')
                sys.stdout.flush()

                counter += 1

        self.log_message('Downloaded {} people and memeberships'.format(counter), fancy=True)

    def grab_person_memberships(self, person_id):
        # this grabs a person and all their memberships

        url = base_url + '/' + person_id + '/'
        r = requests.get(url)
        page_json = json.loads(r.text)

        # save image to disk
        if page_json['image']:
            r = requests.get(page_json['image'], verify=False)
            if r.status_code == 200:
                with open((settings.HEADSHOT_PATH + page_json['id'] + ".jpg"), 'wb') as f:
                    for chunk in r.iter_content(1000):
                        f.write(chunk)
                        f.flush()

        page_json['email'] = None
        for contact_detail in page_json['contact_details']:
            if contact_detail['type'] == 'email':
                if contact_detail['value'] != 'mailto:':
                    page_json['email'] = contact_detail['value']

        page_json['website_url'] = None
        for link in page_json['links']:
            if link['note'] == "web site":
                page_json['website_url'] = link['url']

        return page_json

    def grab_bills(self):

        os.makedirs(self.bills_folder, exist_ok=True)

        if hasattr(settings, 'OCD_CITY_COUNCIL_ID'):
            query_params = {'from_organization__id': settings.OCD_CITY_COUNCIL_ID}
        else:
            query_params = {'from_organization__name': settings.OCD_CITY_COUNCIL_NAME}

        if self.update_since is None:
            max_updated = Bill.objects.all().aggregate(
                Max('ocd_updated_at'))['ocd_updated_at__max']

            if max_updated is None:
                max_updated = datetime.datetime(1900, 1, 1)
        else:
            max_updated = self.update_since

        query_params['sort'] = 'updated_at'
        query_params['updated_at__gte'] = max_updated.isoformat()

        self.log_message('Getting bills since {}'.format(query_params['updated_at__gte']), style='NOTICE')

        search_url = '{}/bills/'.format(base_url)
        search_results = requests.get(search_url, params=query_params)
        page_json = search_results.json()

        counter = 0
        for page_num in range(page_json['meta']['max_page']):

            query_params['page'] = int(page_num) + 1
            result_page = requests.get(search_url, params=query_params)

            for result in result_page.json()['results']:

                bill_url = '{base}/{bill_id}/'.format(
                    base=base_url, bill_id=result['id'])
                bill_detail = requests.get(bill_url)

                bill_json = bill_detail.json()
                ocd_uuid = bill_json['id'].split('/')[-1]
                bill_filename = '{}.json'.format(ocd_uuid)

                with open(os.path.join(self.bills_folder, bill_filename), 'w') as f:
                    f.write(json.dumps(bill_json))

                counter += 1

                print('.', end='')
                sys.stdout.flush()

                if counter % 1000 == 0:
                    print('\n')
                    self.log_message('Downloaded {} bills'.format(counter))

        self.log_message('Downloaded {} bills'.format(counter), fancy=True)

    def grab_events(self):

        os.makedirs(self.events_folder, exist_ok=True)

        events_url = '{0}/events/'.format(base_url)

        params = {'jurisdiction_id': settings.OCD_JURISDICTION_ID}

        if self.update_since is None:
            max_updated = Event.objects.all().aggregate(
                Max('ocd_updated_at'))['ocd_updated_at__max']

            if max_updated is None:
                max_updated = datetime.datetime(1900, 1, 1)
        else:
            max_updated = self.update_since

        params['updated_at__gte'] = max_updated.isoformat()

        r = requests.get(events_url, params=params)
        page_json = json.loads(r.text)

        counter = 0
        for i in range(page_json['meta']['max_page']):

            params['page'] = str(i + 1)
            r = requests.get(events_url, params=params)
            page_json = json.loads(r.text)

            for event in page_json['results']:

                ocd_uuid = event['id'].split('/')[-1]
                event_filename = '{}.json'.format(ocd_uuid)

                event_url = base_url + '/' + event['id'] + '/'
                r = requests.get(event_url)

                if r.status_code == 200:
                    page_json = json.loads(r.text)

                    with open(os.path.join(self.events_folder, event_filename), 'w') as f:
                        f.write(json.dumps(page_json))

                    counter += 1

                    print('.', end='')
                    sys.stdout.flush()

                    if counter % 1000 == 0:
                        print('\n')
                        self.log_message('Downloaded {} events'.format(counter))
                else:
                    self.log_message('Skipping event {} (cannot retrieve event data)'.format(event['id']), style='ERROR')

        self.log_message('Downloaded {} events'.format(counter), fancy=True)

    ###########################
    ###                     ###
    ### INSERT RAW ENTITIES ###
    ###                     ###
    ###########################

    def remake_raw(self, entity_type, delete=False):

        if delete:
            self.executeTransaction(
                'TRUNCATE councilmatic_core_{} CASCADE'.format(entity_type))

            print("deleted all {}".format(entity_type))

        self.executeTransaction('DROP TABLE IF EXISTS raw_{}'.format(entity_type))

        self.executeTransaction('''
            CREATE TABLE raw_{0} AS (
              SELECT * FROM councilmatic_core_{0}
            ) WITH NO DATA
        '''.format(entity_type))

    def setup_raw(self,
                  entity_type,
                  delete=False,
                  pk_cols=['ocd_id'],
                  updated_at=True):

        self.remake_raw(entity_type, delete=delete)

        if pk_cols:
            self.executeTransaction('''
                ALTER TABLE raw_{0} ADD PRIMARY KEY ({1})
            '''.format(entity_type, ','.join(pk_cols)))

        if updated_at:
            self.executeTransaction('''
                ALTER TABLE raw_{}
                ALTER COLUMN updated_at SET DEFAULT NOW()
            '''.format(entity_type))

    def create_legislative_sessions(self):
        session_ids = []

        if hasattr(settings, 'LEGISLATIVE_SESSIONS') and settings.LEGISLATIVE_SESSIONS:
            session_ids = settings.LEGISLATIVE_SESSIONS
        else:
            url = base_url + '/' + settings.OCD_JURISDICTION_ID + '/'
            r = requests.get(url)
            page_json = json.loads(r.text)
            session_ids = [session['identifier']
                           for session in page_json['legislative_sessions']]

        # Sort so most recent session last
        session_ids.sort()
        for leg_session in session_ids:
            obj, created = LegislativeSession.objects.get_or_create(
                identifier=leg_session,
                jurisdiction_ocd_id=settings.OCD_JURISDICTION_ID,
                name='%s Legislative Session' % leg_session,
            )
            if created and DEBUG:
                print('adding legislative session: %s' % obj.name)

    def insert_raw_organizations(self, delete=False):

        self.setup_raw('organization', delete=delete)

        inserts = []

        insert_query = '''
            INSERT INTO raw_organization (
                ocd_id,
                name,
                classification,
                source_url,
                slug,
                parent_id
            ) VALUES (
                :ocd_id,
                :name,
                :classification,
                :source_url,
                :slug,
                :parent_id
            )
            '''

        for organization_json in os.listdir(self.organizations_folder):

            with open(os.path.join(self.organizations_folder, organization_json)) as f:
                org_info = json.loads(f.read())

            source_url = None
            if org_info['sources']:
                source_url = org_info['sources'][0]['url']

            parent_ocd_id = None
            if org_info['parent']:
                parent_ocd_id = org_info['parent']['id']

            ocd_part = org_info['id'].rsplit('-', 1)[1]
            slug = '{0}-{1}'.format(slugify(org_info['name']),ocd_part)

            insert = {
                'ocd_id': org_info['id'],
                'name': org_info['name'],
                'classification': org_info['classification'],
                'source_url': source_url,
                'slug': slug,
                'parent_id': parent_ocd_id,
            }

            inserts.append(insert)

        if inserts:
            self.executeTransaction(sa.text(insert_query), *inserts)

        raw_count = self.connection.execute('select count(*) from raw_organization').first().count

        self.log_message('Inserted {0} raw organizations'.format(raw_count), style='SUCCESS')

    def insert_raw_posts(self, delete=False):

        self.setup_raw('post', delete=delete)

        inserts = []

        insert_query = '''
            INSERT INTO raw_post (
                ocd_id,
                label,
                role,
                organization_id,
                division_ocd_id
            ) VALUES (
                :ocd_id,
                :label,
                :role,
                :organization_id,
                :division_ocd_id
            )
        '''

        for post_json in os.listdir(self.posts_folder):

            with open(os.path.join(self.posts_folder, post_json)) as f:
                post_info = json.loads(f.read())

            insert = {
                'ocd_id': post_info['id'],
                'label': post_info['label'],
                'role': post_info['role'],
                'organization_id': post_info['org_ocd_id'],
                'division_ocd_id': post_info['division_id'],
            }

            inserts.append(insert)

        if inserts:
            self.executeTransaction(sa.text(insert_query), *inserts)

        raw_count = self.connection.execute('select count(*) from raw_post').first().count

        self.log_message('Inserted {0} raw posts'.format(raw_count), style='SUCCESS')

    def insert_raw_people(self, delete=False):

        self.setup_raw('person', delete=delete)

        inserts = []

        insert_query = '''
            INSERT INTO raw_person (
                ocd_id,
                name,
                headshot,
                source_url,
                source_note,
                website_url,
                email,
                slug
            ) VALUES (
                :ocd_id,
                :name,
                :headshot,
                :source_url,
                :source_note,
                :website_url,
                :email,
                :slug
            )
        '''

        for person_json in os.listdir(self.people_folder):

            with open(os.path.join(self.people_folder, person_json)) as f:
                person_info = json.loads(f.read())

            source_url = None
            if person_info['sources']:
                source_url = person_info['sources'][0]['url']

            source_note = None
            if person_info['sources']:
                source_note = person_info['sources'][0]['note']

            ocd_part = person_info['id'].rsplit('-', 1)[1]
            slug = '{0}-{1}'.format(slugify(person_info['name']),ocd_part)

            insert = {
                'ocd_id': person_info['id'],
                'name': person_info['name'],
                'headshot': person_info['image'],
                'source_url': source_url,
                'source_note': source_note,
                'website_url': person_info['website_url'],
                'email': person_info['email'],
                'slug': slug,
            }

            inserts.append(insert)

        if inserts:
            self.executeTransaction(sa.text(insert_query), *inserts)

        raw_count = self.connection.execute('select count(*) from raw_person').first().count

        self.log_message('Inserted {0} raw people\n'.format(raw_count), style='SUCCESS')

    def insert_raw_memberships(self, delete=False):

        self.setup_raw('membership', delete=delete, pk_cols=[])


        inserts = []

        insert_query = '''
            INSERT INTO raw_membership (
                label,
                role,
                start_date,
                end_date,
                organization_id,
                person_id,
                post_id
            ) VALUES (
                :label,
                :role,
                :start_date,
                :end_date,
                :organization_id,
                :person_id,
                :post_id
            )
        '''

        for person_json in os.listdir(self.people_folder):

            with open(os.path.join(self.people_folder, person_json)) as f:
                person_info = json.loads(f.read())

            for membership_json in person_info['memberships']:

                end_date = parse_date(membership_json['end_date'])

                start_date = parse_date(membership_json['start_date'])

                post_id = None
                if membership_json['post']:
                    post_id = membership_json['post']['id']

                insert = {
                    'label': membership_json['label'],
                    'role': membership_json['role'],
                    'start_date': start_date,
                    'end_date': end_date,
                    'organization_id': membership_json['organization']['id'],
                    'person_id': person_info['id'],
                    'post_id': post_id,
                }

                inserts.append(insert)

        if inserts:
            self.executeTransaction(sa.text(insert_query), *inserts)

        raw_count = self.connection.execute('select count(*) from raw_membership').first().count

        self.log_message('Inserted {0} raw memberships\n'.format(raw_count), style='SUCCESS')

    def insert_raw_bills(self, delete=False):

        self.setup_raw('bill', delete=delete)

        inserts = []

        insert_query = '''
            INSERT INTO raw_bill (
                ocd_id,
                ocd_created_at,
                ocd_updated_at,
                description,
                identifier,
                classification,
                source_url,
                source_note,
                from_organization_id,
                full_text,
                ocr_full_text,
                abstract,
                legislative_session_id,
                bill_type,
                subject,
                slug
            ) VALUES (
                :ocd_id,
                :ocd_created_at,
                :ocd_updated_at,
                :description,
                :identifier,
                :classification,
                :source_url,
                :source_note,
                :from_organization_id,
                :full_text,
                :ocr_full_text,
                :abstract,
                :legislative_session_id,
                :bill_type,
                :subject,
                :slug
            )
            '''

        counter = 0

        for bill_json in os.listdir(self.bills_folder):

            with open(os.path.join(self.bills_folder, bill_json)) as f:
                bill_info = json.loads(f.read())

            source_url = None
            for source in bill_info['sources']:
                if source['note'] == 'web':
                    source_url = source['url']

            full_text = None
            if 'full_text' in bill_info['extras']:
                full_text = bill_info['extras']['full_text']

            ocr_full_text = None
            if 'ocr_full_text' in bill_info['extras']:
                ocr_full_text = bill_info['extras']['ocr_full_text']

            elif 'plain_text' in bill_info['extras']:
                ocr_full_text = bill_info['extras']['plain_text']

            abstract = None
            if bill_info['abstracts']:
                abstract = bill_info['abstracts'][0]['abstract']

            if bill_info['extras'].get('local_classification'):
                bill_type = bill_info['extras']['local_classification']

            elif len(bill_info['classification']) == 1:
                bill_type = bill_info['classification'][0]

            else:
                raise Exception(bill_info['classification'])

            subject = None
            if 'subject' in bill_info and bill_info['subject']:
                subject = bill_info['subject'][0]


            ocd_part = bill_info['id'].rsplit('-', 1)[1]
            slug = '{0}-{1}'.format(slugify(bill_info['identifier']),ocd_part)

            insert = {
                'ocd_id': bill_info['id'],
                'ocd_created_at': bill_info['created_at'],
                'ocd_updated_at': bill_info['updated_at'],
                'description': bill_info['title'],
                'identifier': bill_info['identifier'],
                'classification': bill_info['classification'][0],
                'source_url': source_url,
                'source_note': bill_info['sources'][0]['note'],
                'from_organization_id': bill_info['from_organization']['id'],
                'full_text': full_text,
                'ocr_full_text': ocr_full_text,
                'abstract': abstract,
                'legislative_session_id': bill_info['legislative_session']['identifier'],
                'bill_type': bill_type,
                'subject': subject,
                'slug': slug,
            }

            inserts.append(insert)

            if inserts and len(inserts) % 10000 == 0:
                self.executeTransaction(sa.text(insert_query), *inserts)

                counter += 10000
                self.log_message('Inserted {} raw bills'.format(counter))

                inserts = []

        if inserts:

            self.executeTransaction(sa.text(insert_query), *inserts)

            counter += len(inserts)

        self.log_message('Inserted a total of {} raw bills\n'.format(counter), style='SUCCESS')

    def insert_raw_actions(self, delete=False):

        pk_cols = ['bill_id', '"order"']

        self.setup_raw('action', delete=delete, pk_cols=pk_cols)

        inserts = []

        insert_query = '''
            INSERT INTO raw_action (
                date,
                classification,
                description,
                organization_id,
                bill_id,
                "order"
            ) VALUES (
                :date,
                :classification,
                :description,
                :organization_id,
                :bill_id,
                :order
            )
            '''

        counter = 0
        for bill_json in os.listdir(self.bills_folder):

            with open(os.path.join(self.bills_folder, bill_json)) as f:
                bill_info = json.loads(f.read())

            for order, action in enumerate(bill_info['actions']):

                classification = None
                if action['classification']:
                    classification = action['classification'][0]

                action_date = app_timezone.localize(date_parser.parse(action['date']))

                insert = {
                    'date': action_date,
                    'classification': classification,
                    'description': action['description'],
                    'organization_id': action['organization']['id'],
                    'bill_id': bill_info['id'],
                    'order': order
                }

                inserts.append(insert)

                if inserts and len(inserts) % 10000 == 0:
                    self.executeTransaction(sa.text(insert_query), *inserts)

                    counter += 10000

                    self.log_message('Inserted {0} actions'.format(counter))

                    inserts = []

        if inserts:
            self.executeTransaction(sa.text(insert_query), *inserts)

            counter += len(inserts)

        self.log_message('Inserted {0} actions\n'.format(counter), style='SUCCESS')


    def insert_raw_action_related_entity(self, delete=False):

        self.remake_raw('actionrelatedentity', delete=delete)

        self.executeTransaction('''
            ALTER TABLE raw_actionrelatedentity
            ALTER COLUMN updated_at SET DEFAULT NOW()
        ''')

        inserts = []

        insert_query = '''
            INSERT INTO raw_actionrelatedentity (
                entity_type,
                entity_name,
                organization_ocd_id,
                person_ocd_id,
                action_id
            ) VALUES (
                :entity_type,
                :entity_name,
                :organization_ocd_id,
                :person_ocd_id,
                :action_id
            )
            '''

        read_connection = engine.connect()
        actions = read_connection.execute('''
            SELECT id, bill_id, "order" FROM councilmatic_core_action
        ''')

        counter = 0
        for action in actions:

            action_id, bill_id, order = action

            ocd_uuid = bill_id.split('/')[-1]
            bill_filename = '{}.json'.format(ocd_uuid)

            with open(os.path.join(self.bills_folder, bill_filename)) as f:
                bill_info = json.loads(f.read())

            action = bill_info['actions'][order]

            for related_entity in action['related_entities']:

                person_id = None
                organization_id = None

                if related_entity['entity_type'] == 'organization':

                    organization_id = related_entity['organization_id']

                    if not organization_id:
                        org_id = self.connection.execute(sa.text("""
                            SELECT ocd_id
                            FROM councilmatic_core_organization
                            WHERE name = :name
                            LIMIT 1
                        """), name=related_entity['name']).first()

                        if org_id:
                            organization_id = org_id.ocd_id

                elif related_entity['entity_type'] == 'person':

                    person_id = related_entity['person_id']

                    if not person_id:
                        person_id = self.connection.execute(sa.text("""
                            SELECT ocd_id
                            FROM councilmatic_core_person
                            WHERE name = :name
                            LIMIT 1
                        """), name=related_entity['name']).first()

                        if person_id:
                            person_id = person_id.ocd_id

                insert = {
                    'entity_type': related_entity['entity_type'],
                    'entity_name': related_entity['name'],
                    'organization_ocd_id': organization_id,
                    'person_ocd_id': person_id,
                    'action_id': action_id,
                }

                inserts.append(insert)

                if inserts and len(inserts) % 10000 == 0:
                    self.executeTransaction(sa.text(insert_query), *inserts)

                    counter += 10000
                    self.log_message('Inserted {0} action related entities'.format(counter))

                    inserts = []

        if inserts:
            self.executeTransaction(sa.text(insert_query), *inserts)

            counter += len(inserts)

        self.log_message('Inserted {0} action related entities\n'.format(counter), style='SUCCESS')


    def insert_raw_sponsorships(self, delete=False):

        self.remake_raw('sponsorship', delete=delete)

        self.executeTransaction('''
                ALTER TABLE raw_sponsorship
                ALTER COLUMN updated_at SET DEFAULT NOW()
            ''')

        self.executeTransaction('''
                DROP TABLE IF EXISTS raw_sponsorship_temp
            ''')

        self.executeTransaction('''
                CREATE TABLE raw_sponsorship_temp AS (
                    SELECT * FROM councilmatic_core_sponsorship
                ) WITH NO DATA
            ''')

        self.executeTransaction('''
                ALTER TABLE raw_sponsorship_temp
                ALTER COLUMN updated_at SET DEFAULT NOW()
            ''')

        inserts = []

        insert_query = '''
            INSERT INTO raw_sponsorship_temp (
                classification,
                is_primary,
                bill_id,
                person_id
            ) VALUES (
                :classification,
                :is_primary,
                :bill_id,
                :person_id
            )
            '''

        counter = 0
        for bill_json in os.listdir(self.bills_folder):

            with open(os.path.join(self.bills_folder, bill_json)) as f:
                bill_info = json.loads(f.read())

            for sponsorship in bill_info['sponsorships']:

                # TODO: Sponsorships can also be organizations but we're
                # waiting to see what that actually means

                if sponsorship['entity_type'] == 'person':

                    insert = {
                        'classification': sponsorship['classification'],
                        'is_primary': sponsorship['primary'],
                        'bill_id': bill_info['id'],
                        'person_id': sponsorship['entity_id'],
                    }

                    inserts.append(insert)

                if inserts and len(inserts) % 10000 == 0:
                    self.executeTransaction(sa.text(insert_query), *inserts)

                    counter += 10000
                    self.log_message('Inserted {0} raw sponsorships'.format(counter))

                    inserts = []

        if inserts:
            self.executeTransaction(sa.text(insert_query), *inserts)

        # Temproary measure to make sure we can actually update things.
        self.executeTransaction('''
                INSERT INTO raw_sponsorship
                  SELECT DISTINCT ON (classification, is_primary, bill_id, person_id)
                  *
                  FROM raw_sponsorship_temp
            ''')

        raw_count = self.connection.execute('select count(*) from raw_sponsorship').first().count
        self.log_message('Inserted {0} raw sponsorships\n'.format(raw_count), style='SUCCESS')

    def insert_raw_billdocuments(self, delete=False):

        self.remake_raw('billdocument', delete=delete)

        self.executeTransaction('''
                ALTER TABLE raw_billdocument
                ALTER COLUMN updated_at SET DEFAULT NOW()
            ''')

        inserts = []

        insert_query = '''
            INSERT INTO raw_billdocument (
                note,
                url,
                bill_id,
                document_type
            ) VALUES (
                :note,
                :url,
                :bill_id,
                :document_type
            )
        '''

        counter = 0
        for bill_json in os.listdir(self.bills_folder):

            with open(os.path.join(self.bills_folder, bill_json)) as f:
                bill_info = json.loads(f.read())

            for document in bill_info['documents']:

                insert = {
                    'note': document['note'],
                    'url': document['links'][0]['url'],
                    'bill_id': bill_info['id'],
                    'document_type': 'A',
                }

                inserts.append(insert)

            for document in bill_info['versions']:

                insert = {
                    'note': document['note'],
                    'url': document['links'][0]['url'],
                    'bill_id': bill_info['id'],
                    'document_type': 'V',
                }

                inserts.append(insert)

            if inserts and len(inserts) % 10000 == 0:
                self.executeTransaction(sa.text(insert_query), *inserts)

                counter += 10000

                self.log_message('Inserted {0} raw bill attachments and versions'.format(counter))

                inserts = []

        if inserts:
            self.executeTransaction(sa.text(insert_query), *inserts)

            counter += len(inserts)

        self.log_message('Inserted {0} raw bill attachments and versions\n'.format(counter), style='SUCCESS')


    def insert_raw_events(self, delete=False):
        self.setup_raw('event', delete=delete)

        inserts = []

        insert_query = '''
            INSERT INTO raw_event (
                ocd_id,
                ocd_created_at,
                ocd_updated_at,
                name,
                description,
                classification,
                start_time,
                end_time,
                all_day,
                status,
                location_name,
                location_url,
                source_url,
                source_note,
                slug
            ) VALUES (
                :ocd_id,
                :ocd_created_at,
                :ocd_updated_at,
                :name,
                :description,
                :classification,
                :start_time,
                :end_time,
                :all_day,
                :status,
                :location_name,
                :location_url,
                :source_url,
                :source_note,
                :slug
            )
        '''

        counter = 0

        for event_json in os.listdir(self.events_folder):

            with open(os.path.join(self.events_folder, event_json)) as f:
                event_info = json.loads(f.read())

            ocd_id = event_info['id']

            truncator = Truncator(event_info['name'])
            ocd_part = ocd_id.rsplit('-', 1)[1]
            slug = '{0}-{1}'.format(slugify(truncator.words(5)), ocd_part)

            insert = {
                'ocd_id': ocd_id,
                'ocd_created_at': event_info['created_at'],
                'ocd_updated_at': event_info['updated_at'],
                'name': event_info['name'],
                'description': event_info['description'],
                'classification': event_info['classification'],
                'start_time': parse_datetime(event_info['start_time']),
                'end_time': parse_datetime(event_info['end_time']) if event_info['end_time'] else None,
                'all_day': event_info['all_day'],
                'status': event_info['status'],
                'location_name': event_info['location']['name'],
                'location_url': event_info['location']['url'],
                'source_url': event_info['sources'][0]['url'],
                'source_note': event_info['sources'][0]['note'],
                'slug': slug,
            }

            inserts.append(insert)

            if inserts and len(inserts) % 10000 == 0:
                self.executeTransaction(sa.text(insert_query), *inserts)

                counter += 10000

                self.log_message('Inserted {0} raw events'.format(counter))

                inserts = []

        if inserts:
            self.executeTransaction(sa.text(insert_query), *inserts)

            counter += len(inserts)

        self.log_message('Inserted {0} raw events\n'.format(counter), style='SUCCESS')

    def insert_raw_eventparticipants(self, delete=False):
        pk_cols = ['event_id', 'entity_name', 'entity_type']

        self.setup_raw('eventparticipant',
                       delete=delete,
                       pk_cols=pk_cols)

        inserts = []

        insert_query = '''
            INSERT INTO raw_eventparticipant (
                note,
                entity_name,
                entity_type,
                event_id
            ) VALUES (
                :note,
                :entity_name,
                :entity_type,
                :event_id
            )
            '''

        counter = 0
        for event_json in os.listdir(self.events_folder):

            with open(os.path.join(self.events_folder, event_json)) as f:
                event_info = json.loads(f.read())

            for participant in event_info['participants']:

                insert = {
                    'note': participant['note'],
                    'entity_name': participant['entity_name'],
                    'entity_type': participant['entity_type'],
                    'event_id': event_info['id'],
                }

                inserts.append(insert)

                if inserts and len(inserts) % 10000 == 0:
                    self.executeTransaction(sa.text(insert_query), *inserts)

                    counter += 10000

                    self.log_message('Inserted {} raw event participants'.format(counter))

                    inserts = []

        if inserts:
            self.executeTransaction(sa.text(insert_query), *inserts)
            counter += len(inserts)

        self.log_message('Inserted {0} event participants\n'.format(counter), style='SUCCESS')

    def insert_raw_eventdocuments(self, delete=False):
        pk_cols = ['event_id', 'url']

        self.setup_raw('eventdocument',
                       delete=delete,
                       pk_cols=pk_cols,
                       updated_at=False)

        inserts = []

        insert_query = '''
            INSERT INTO raw_eventdocument (
                event_id,
                note,
                url
            ) VALUES (
                :event_id,
                :note,
                :url
            )
            '''

        counter = 0
        for event_json in os.listdir(self.events_folder):

            with open(os.path.join(self.events_folder, event_json)) as f:
                event_info = json.loads(f.read())

            for document in event_info['documents']:

                insert = {
                    'event_id': event_info['id'],
                    'note': document['note'],
                    'url': document['links'][0]['url'],
                }

                inserts.append(insert)

                if inserts and len(inserts) % 10000 == 0:
                    self.executeTransaction(sa.text(insert_query), *inserts)

                    counter += 10000

                    self.log_message('Inserted {} raw event documents'.format(counter))

                    inserts = []

        if inserts:
            self.executeTransaction(sa.text(insert_query), *inserts)

            counter += len(inserts)

        self.log_message('Inserted {0} event documents\n'.format(counter), style='SUCCESS')

    def insert_raw_event_agenda_items(self, delete=False):
        pk_cols = ['event_id', '"order"']

        self.setup_raw('eventagendaitem',
                       delete=delete,
                       pk_cols=pk_cols)

        inserts = []

        insert_query = '''
            INSERT INTO raw_eventagendaitem (
                "order",
                description,
                event_id,
                bill_id,
                note
            ) VALUES (
                :order,
                :description,
                :event_id,
                :bill_id,
                :note

            )
            '''

        counter = 0

        for event_json in os.listdir(self.events_folder):

            with open(os.path.join(self.events_folder, event_json)) as f:
                event_info = json.loads(f.read())

            for item in event_info['agenda']:

                bill_id = None
                note = None

                try:
                    related_entity = item['related_entities'][0]

                    # Only capture related entities when they are bills

                    if related_entity['entity_type'] == 'bill':
                        bill_id = related_entity['entity_id']
                        note = related_entity['note']

                except IndexError:
                    pass

                insert = {
                    'order': item['order'],
                    'description': item['description'],
                    'event_id': event_info['id'],
                    'bill_id': bill_id,
                    'note': note,
                }

                inserts.append(insert)
                if inserts and len(inserts) % 10000 == 0:
                    self.executeTransaction(sa.text(insert_query), *inserts)

                    counter += 10000

                    self.log_message('Inserted {} raw event agenda items'.format(counter))

                    inserts = []

        if inserts:
            self.executeTransaction(sa.text(insert_query), *inserts)

            counter += len(inserts)

        self.log_message('Inserted {0} raw event agenda items\n'.format(counter), style='SUCCESS')

    ################################
    ###                          ###
    ### UPDATE EXISTING ENTITIES ###
    ###                          ###
    ################################

    def setup_update(self, entity_type):
        self.executeTransaction('DROP TABLE IF EXISTS change_{}'.format(entity_type))

        self.executeTransaction('''
            CREATE TABLE change_{} (
                ocd_id VARCHAR,
                PRIMARY KEY (ocd_id)
            )
        '''.format(entity_type))

    def get_update_parts(self, cols, extra_cols):
        wheres = []
        sets = []
        fields = []
        for col in cols:
            condition = '''
                ((raw."{0}" IS NOT NULL OR dat."{0}" IS NOT NULL) AND
                COALESCE(raw."{0}"::VARCHAR, '') <> COALESCE(dat."{0}"::VARCHAR, ''))
            '''.format(col)
            wheres.append(condition)

            sets.append('"{0}"=s."{0}"'.format(col))
            fields.append('raw."{0}"'.format(col))

        for col in extra_cols:
            sets.append('"{0}"=s."{0}"'.format(col))
            fields.append('raw."{0}"'.format(col))

        where_clause = ' OR '.join(wheres)
        set_values = ', '.join(sets)
        fields = ', '.join(fields)

        return where_clause, set_values, fields

    def update_entity_type(self, entity_type, cols=[], extra_cols=['updated_at']):

        self.setup_update(entity_type)

        where_clause, set_values, fields = self.get_update_parts(cols, extra_cols)

        find_changes = '''
            INSERT INTO change_{0}
              SELECT raw.ocd_id
              FROM raw_{0} AS raw
              JOIN councilmatic_core_{0} AS dat
                ON raw.ocd_id = dat.ocd_id
              WHERE {1}
        '''.format(entity_type, where_clause)

        update_dat = '''
            UPDATE councilmatic_core_{entity_type} SET
              {set_values}
            FROM (
              SELECT
                {fields}
              FROM raw_{entity_type} AS raw
              JOIN change_{entity_type} AS change
                ON raw.ocd_id = change.ocd_id
            ) AS s
            WHERE councilmatic_core_{entity_type}.ocd_id = s.ocd_id
        '''.format(set_values=set_values,
                   fields=fields,
                   entity_type=entity_type)

        self.executeTransaction(find_changes)
        self.executeTransaction(update_dat)

        change_count = self.connection.execute('select count(*) from change_{}'.format(entity_type)).first().count

        self.log_message('Found {0} changed {1}'.format(change_count, entity_type), style='SUCCESS')

    def update_existing_organizations(self):

        cols = [
            'ocd_id',
            'name',
            'classification',
            'source_url',
            'parent_id',
            'slug'
        ]

        self.update_entity_type('organization', cols=cols)

    def update_existing_posts(self):

        cols = [
            'ocd_id',
            'label',
            'role',
            'organization_id',
            'division_ocd_id'
        ]

        self.update_entity_type('post', cols=cols)

    def update_existing_people(self):
        cols = [
            'ocd_id',
            'name',
            'headshot',
            'source_url',
            'source_note',
            'website_url',
            'email',
            'slug',
        ]
        self.update_entity_type('person', cols=cols)

    def update_existing_memberships(self):

        self.executeTransaction('DROP TABLE IF EXISTS change_membership')
        self.executeTransaction('''
            CREATE TABLE change_membership (
                organization_id VARCHAR,
                person_id VARCHAR,
                post_id VARCHAR,
                start_date DATE,
                end_date DATE
            )
        ''')

        cols = [
           'label',
           'role',
           'start_date',
           'end_date',
           'organization_id',
           'person_id',
           'post_id'
        ]

        where_clause, set_values, fields = self.get_update_parts(cols, [])

        find_changes = '''
            INSERT INTO change_membership
              SELECT
                raw.organization_id,
                raw.person_id,
                raw.post_id,
                raw.start_date,
                raw.end_date
              FROM raw_membership AS raw
              JOIN councilmatic_core_membership AS dat
                ON (raw.organization_id = dat.organization_id
                    AND raw.person_id = dat.person_id
                    AND COALESCE(raw.post_id, '') = COALESCE(dat.post_id, '')
                    AND COALESCE(raw.start_date, '1900-01-01') = COALESCE(dat.start_date, '1900-01-01')
                    AND COALESCE(raw.end_date, NOW()) = COALESCE(dat.end_date, NOW()))
              WHERE {}
        '''.format(where_clause)

        update_dat = '''
            UPDATE councilmatic_core_membership SET
              {set_values}
            FROM (
              SELECT
                {fields}
              FROM raw_membership AS raw
              JOIN change_membership AS change
                ON (raw.organization_id = change.organization_id
                    AND raw.person_id = change.person_id
                    AND COALESCE(raw.post_id, '') = COALESCE(change.post_id, '')
                    AND COALESCE(raw.start_date, '1900-01-01') = COALESCE(change.start_date, '1900-01-01')
                    AND COALESCE(raw.end_date, NOW()) = COALESCE(change.end_date, NOW()))
            ) AS s
            WHERE councilmatic_core_membership.organization_id = s.organization_id
              AND councilmatic_core_membership.person_id = s.person_id
              AND COALESCE(councilmatic_core_membership.post_id, '') = COALESCE(s.post_id, '')
              AND COALESCE(councilmatic_core_membership.start_date, '1900-01-01') = COALESCE(s.start_date, '1900-01-01')
              AND COALESCE(councilmatic_core_membership.end_date, NOW()) = COALESCE(s.end_date, NOW())
        '''.format(set_values=set_values,
                   fields=fields)

        self.executeTransaction(find_changes)
        self.executeTransaction(update_dat)

        change_count = self.connection.execute('select count(*) from change_membership').first().count

        self.log_message('Found {0} changed membership'.format(change_count), style='SUCCESS')

    def update_existing_bills(self):
        cols = [
            'ocd_id',
            'ocd_created_at',
            'ocd_updated_at',
            'description',
            'identifier',
            'bill_type',
            'classification',
            'source_url',
            'source_note',
            'subject',
            'from_organization_id',
            'full_text',
            'ocr_full_text',
            'abstract',
            'last_action_date',
            'legislative_session_id',
            'slug',
        ]
        self.update_entity_type('bill', cols=cols)

    def update_existing_actions(self):

        self.executeTransaction('DROP TABLE IF EXISTS change_action')
        self.executeTransaction('''
            CREATE TABLE change_action (
                bill_id VARCHAR,
                "order" INTEGER
            )
        ''')

        cols = [
            'date',
            'classification',
            'description',
            'order',
            'bill_id',
            'organization_id',
        ]

        where_clause, set_values, fields = self.get_update_parts(cols, ['updated_at'])

        find_changes = '''
            INSERT INTO change_action
              SELECT
                raw.bill_id,
                raw."order"
              FROM raw_action AS raw
              JOIN councilmatic_core_action AS dat
                ON (raw.bill_id = dat.bill_id
                    AND raw."order" = dat."order")
              WHERE {}
        '''.format(where_clause)

        update_dat = '''
            UPDATE councilmatic_core_action SET
              {set_values}
            FROM (
              SELECT
                {fields}
              FROM raw_action AS raw
              JOIN change_action AS change
                ON (raw.bill_id = change.bill_id
                    AND raw."order" = change."order")
            ) AS s
            WHERE councilmatic_core_action.bill_id = s.bill_id
              AND councilmatic_core_action."order" = s."order"
        '''.format(set_values=set_values,
                   fields=fields)

        self.executeTransaction(find_changes)

        self.executeTransaction(update_dat)
        change_count = self.connection.execute('select count(*) from change_action').first().count

        self.log_message('Found {0} changed action'.format(change_count), style='SUCCESS')

    def update_existing_action_related_entity(self):
        self.executeTransaction('DROP TABLE IF EXISTS change_actionrelatedentity')
        self.executeTransaction('''
            CREATE TABLE change_actionrelatedentity (
                organization_ocd_id VARCHAR,
                person_ocd_id VARCHAR,
                action_id INTEGER
            )
        ''')

        cols = [
            'entity_type',
            'entity_name',
            'organization_ocd_id',
            'person_ocd_id',
            'action_id',
        ]

        where_clause, set_values, fields = self.get_update_parts(cols, ['updated_at'])

        find_changes = '''
            INSERT INTO change_actionrelatedentity
              SELECT
                raw.organization_ocd_id,
                raw.person_ocd_id,
                raw.action_id
              FROM raw_actionrelatedentity AS raw
              JOIN councilmatic_core_actionrelatedentity AS dat
                ON (COALESCE(raw.organization_ocd_id, '') = COALESCE(dat.organization_ocd_id, '')
                    AND COALESCE(raw.person_ocd_id, '') = COALESCE(dat.person_ocd_id, '')
                    AND raw.action_id = dat.action_id)
              WHERE {}
        '''.format(where_clause)

        update_dat = '''
            UPDATE councilmatic_core_actionrelatedentity SET
              {set_values}
            FROM (
              SELECT
                {fields}
              FROM raw_actionrelatedentity AS raw
              JOIN change_actionrelatedentity AS change
                ON (COALESCE(raw.organization_ocd_id, '') = COALESCE(change.organization_ocd_id, '')
                    AND COALESCE(raw.person_ocd_id, '') = COALESCE(change.person_ocd_id, '')
                    AND raw.action_id = change.action_id)
            ) AS s
            WHERE COALESCE(councilmatic_core_actionrelatedentity.organization_ocd_id, '') = COALESCE(s.organization_ocd_id, '')
              AND COALESCE(councilmatic_core_actionrelatedentity.person_ocd_id, '') = COALESCE(s.person_ocd_id, '')
              AND councilmatic_core_actionrelatedentity.action_id = s.action_id
        '''.format(set_values=set_values,
                   fields=fields)

        self.executeTransaction(find_changes)
        self.executeTransaction(update_dat)

        change_count = self.connection.execute('select count(*) from change_actionrelatedentity').first().count

        self.log_message('Found {0} changed action related entities'.format(change_count), style='SUCCESS')

    def update_existing_sponsorships(self):
        self.executeTransaction('DROP TABLE IF EXISTS change_sponsorship')
        self.executeTransaction('''
            CREATE TABLE change_sponsorship (
                classification VARCHAR,
                is_primary BOOLEAN,
                bill_id VARCHAR,
                person_id VARCHAR
            )
        ''')

        cols = [
            'classification',
            'is_primary',
            'bill_id',
            'person_id',
        ]

        where_clause, set_values, fields = self.get_update_parts(cols, ['updated_at'])

        find_changes = '''
            INSERT INTO change_sponsorship
              SELECT
                raw.classification,
                raw.is_primary,
                raw.bill_id,
                raw.person_id
              FROM raw_sponsorship AS raw
              JOIN councilmatic_core_sponsorship AS dat
                ON (raw.classification = dat.classification
                    AND raw.is_primary = dat.is_primary
                    AND raw.bill_id = dat.bill_id
                    AND raw.person_id = dat.person_id)
              WHERE {}
        '''.format(where_clause)

        update_dat = '''
            UPDATE councilmatic_core_sponsorship SET
              {set_values}
            FROM (
              SELECT
                {fields}
              FROM raw_sponsorship AS raw
              JOIN change_sponsorship AS change
                ON (raw.classification = change.classification
                    AND raw.is_primary = change.is_primary
                    AND raw.bill_id = change.bill_id
                    AND raw.person_id = change.person_id)
            ) AS s
            WHERE councilmatic_core_sponsorship.classification = s.classification
              AND councilmatic_core_sponsorship.is_primary = s.is_primary
              AND councilmatic_core_sponsorship.bill_id = s.bill_id
              AND councilmatic_core_sponsorship.person_id = s.person_id
        '''.format(set_values=set_values,
                   fields=fields)

        self.executeTransaction(find_changes)

        self.executeTransaction(update_dat)

        change_count = self.connection.execute('select count(*) from change_sponsorship').first().count

        self.log_message('Found {0} changed sponsorships'.format(change_count), style='SUCCESS')

    def update_existing_billdocuments(self):
        self.executeTransaction('DROP TABLE IF EXISTS change_billdocument')
        self.executeTransaction('''
            CREATE TABLE change_billdocument (
                bill_id VARCHAR,
                url VARCHAR,
                document_type VARCHAR
            )
        ''')

        cols = [
            'bill_id',
            'url',
            'document_type',
            'note',
        ]

        where_clause, set_values, fields = self.get_update_parts(cols, ['updated_at'])

        find_changes = '''
            INSERT INTO change_billdocument
              SELECT
                raw.bill_id,
                raw.url,
                raw.document_type
              FROM raw_billdocument AS raw
              JOIN councilmatic_core_billdocument AS dat
                ON (raw.bill_id = dat.bill_id
                    AND raw.url = dat.url
                    AND raw.document_type = dat.document_type)
              WHERE {}
        '''.format(where_clause)

        update_dat = '''
            UPDATE councilmatic_core_billdocument SET
              {set_values}
            FROM (
              SELECT
                {fields}
              FROM raw_billdocument AS raw
              JOIN change_billdocument AS change
                ON (raw.bill_id = change.bill_id
                    AND raw.url = change.url
                    AND raw.document_type = change.document_type)
            ) AS s
            WHERE councilmatic_core_billdocument.bill_id = s.bill_id
              AND councilmatic_core_billdocument.url = s.url
              AND councilmatic_core_billdocument.document_type = s.document_type
        '''.format(set_values=set_values,
                   fields=fields)

        self.executeTransaction(find_changes)

        self.executeTransaction(update_dat)

        change_count = self.connection.execute('select count(*) from change_billdocument').first().count

        self.log_message('Found {0} changed bill documents'.format(change_count), style='SUCCESS')

    def update_existing_events(self):
        cols = [
            'ocd_id',
            'ocd_created_at',
            'ocd_updated_at',
            'name',
            'description',
            'classification',
            'start_time',
            'end_time',
            'all_day',
            'status',
            'location_name',
            'location_url',
            'source_url',
            'source_note',
            'slug',
        ]
        self.update_entity_type('event', cols=cols)

    def update_existing_eventparticipants(self):
        self.executeTransaction('DROP TABLE IF EXISTS change_eventparticipant')
        self.executeTransaction('''
            CREATE TABLE change_eventparticipant (
                event_id VARCHAR,
                entity_name VARCHAR,
                entity_type VARCHAR
            )
        ''')

        cols = [
            'event_id',
            'entity_type',
            'entity_name',
            'note',
        ]

        where_clause, set_values, fields = self.get_update_parts(cols, ['updated_at'])

        find_changes = '''
            INSERT INTO change_eventparticipant
              SELECT
                raw.event_id,
                raw.entity_type,
                raw.entity_name
              FROM raw_eventparticipant AS raw
              JOIN councilmatic_core_eventparticipant AS dat
                ON (raw.event_id = dat.event_id
                    AND raw.entity_type = dat.entity_type
                    AND raw.entity_name = dat.entity_name)
              WHERE {}
        '''.format(where_clause)

        update_dat = '''
            UPDATE councilmatic_core_eventparticipant SET
              {set_values}
            FROM (
              SELECT
                {fields}
              FROM raw_eventparticipant AS raw
              JOIN change_eventparticipant AS change
                ON (raw.event_id = change.event_id
                    AND raw.entity_type = change.entity_type
                    AND raw.entity_name = change.entity_name)
            ) AS s
            WHERE councilmatic_core_eventparticipant.event_id = s.event_id
              AND councilmatic_core_eventparticipant.entity_type = s.entity_type
              AND councilmatic_core_eventparticipant.entity_name = s.entity_name
        '''.format(set_values=set_values,
                   fields=fields)

        self.executeTransaction(find_changes)
        self.executeTransaction(update_dat)

        change_count = self.connection.execute('select count(*) from change_eventparticipant').first().count

        self.log_message('Found {0} changed event participants'.format(change_count), style='SUCCESS')

    def update_existing_eventdocuments(self):
        self.executeTransaction('DROP TABLE IF EXISTS change_eventdocument')
        self.executeTransaction('''
            CREATE TABLE change_eventdocument (
                event_id VARCHAR,
                url VARCHAR
            )
        ''')

        cols = [
            'event_id',
            'url',
            'full_text',
            'note'
        ]

        where_clause, set_values, fields = self.get_update_parts(cols, [])

        find_changes = '''
            INSERT INTO change_eventdocument
              SELECT
                raw.event_id,
                raw.url
              FROM raw_eventdocument AS raw
              JOIN councilmatic_core_eventdocument AS dat
                ON (raw.event_id = dat.event_id
                    AND raw.url = dat.url)
              WHERE {}
        '''.format(where_clause)

        update_dat = '''
            UPDATE councilmatic_core_eventdocument SET
              {set_values}
            FROM (
              SELECT
                {fields}
              FROM raw_eventdocument AS raw
              JOIN change_eventdocument AS change
                ON (raw.event_id = change.event_id
                    AND raw.url = change.url)
            ) AS s
            WHERE councilmatic_core_eventdocument.event_id = s.event_id
              AND councilmatic_core_eventdocument.url = s.url
        '''.format(set_values=set_values,
                   fields=fields)

        self.executeTransaction(find_changes)
        self.executeTransaction(update_dat)

        change_count = self.connection.execute('select count(*) from change_eventdocument').first().count

        self.log_message('Found {0} changed event documents'.format(change_count), style='SUCCESS')

    def update_existing_event_agenda_items(self):
        self.executeTransaction('DROP TABLE IF EXISTS change_eventagendaitem')
        self.executeTransaction('''
            CREATE TABLE change_eventagendaitem (
                event_id VARCHAR,
                "order" INTEGER
            )
        ''')

        cols = [
            'order',
            'description',
            'event_id',
            'bill_id',
            'note',
        ]

        where_clause, set_values, fields = self.get_update_parts(cols, ['updated_at'])

        find_changes = '''
            INSERT INTO change_eventagendaitem
              SELECT
                raw.event_id,
                raw."order"
              FROM raw_eventagendaitem AS raw
              JOIN councilmatic_core_eventagendaitem AS dat
                ON (raw.event_id = dat.event_id
                    AND raw."order" = dat."order")
              WHERE {}
        '''.format(where_clause)

        update_dat = '''
            UPDATE councilmatic_core_eventagendaitem SET
              {set_values}
            FROM (
              SELECT
                {fields}
              FROM raw_eventagendaitem AS raw
              JOIN change_eventagendaitem AS change
                ON (raw.event_id = change.event_id
                    AND raw."order" = change."order")
            ) AS s
            WHERE councilmatic_core_eventagendaitem.event_id = s.event_id
              AND councilmatic_core_eventagendaitem."order" = s."order"
        '''.format(set_values=set_values,
                   fields=fields)

        self.executeTransaction(find_changes)
        self.executeTransaction(update_dat)

        change_count = self.connection.execute('select count(*) from change_eventagendaitem').first().count

        self.log_message('Found {0} changed event agenda items'.format(change_count), style='SUCCESS')

    ########################
    ###                  ###
    ### ADD NEW ENTITIES ###
    ###                  ###
    ########################

    def add_entity_type(self, entity_type, cols=[], extra_cols=['updated_at']):
        self.executeTransaction('DROP TABLE IF EXISTS new_{}'.format(entity_type))
        self.executeTransaction('''
            CREATE TABLE new_{} (
                ocd_id VARCHAR,
                PRIMARY KEY (ocd_id)
            )
        '''.format(entity_type))

        find_new = '''
            INSERT INTO new_{0}
              SELECT raw.ocd_id
              FROM raw_{0} AS raw
              LEFT JOIN councilmatic_core_{0} AS dat
                ON raw.ocd_id = dat.ocd_id
              WHERE dat.ocd_id IS NULL
        '''.format(entity_type)

        self.executeTransaction(find_new)

        cols = cols + extra_cols

        insert_fields = ', '.join(c for c in cols)
        select_fields = ', '.join('raw.{}'.format(c) for c in cols)

        insert_new = '''
            INSERT INTO councilmatic_core_{entity_type} (
              {insert_fields}
            )
              SELECT {select_fields}
              FROM raw_{entity_type} AS raw
              JOIN new_{entity_type} AS new
                USING(ocd_id)
        '''.format(entity_type=entity_type,
                   insert_fields=insert_fields,
                   select_fields=select_fields)

        self.executeTransaction(insert_new)

        new_count = self.connection.execute('select count(*) from new_{}'.format(entity_type)).first().count

        self.log_message('Found {0} new {1}'.format(new_count, entity_type), style='SUCCESS')

    def add_new_organizations(self):
        cols = [
            'ocd_id',
            'name',
            'classification',
            'source_url',
            'parent_id',
            'slug',
        ]

        self.add_entity_type('organization', cols=cols)

    def add_new_posts(self):
        cols = [
            'ocd_id',
            'label',
            'role',
            'organization_id',
            'division_ocd_id'
        ]

        self.add_entity_type('post', cols=cols)

    def add_new_people(self):
        cols = [
            'ocd_id',
            'name',
            'headshot',
            'source_url',
            'source_note',
            'website_url',
            'email',
            'slug',
        ]

        self.add_entity_type('person', cols=cols)

    def add_new_memberships(self):
        self.executeTransaction('DROP TABLE IF EXISTS new_membership')
        self.executeTransaction('''
            CREATE TABLE new_membership (
                organization_id VARCHAR,
                person_id VARCHAR,
                post_id VARCHAR,
                start_date DATE,
                end_date DATE
            )
        ''')

        find_new = '''
            INSERT INTO new_membership
              SELECT
                raw.organization_id,
                raw.person_id,
                raw.post_id,
                raw.start_date,
                raw.end_date
              FROM raw_membership AS raw
              LEFT JOIN councilmatic_core_membership AS dat
                ON (raw.organization_id = dat.organization_id
                    AND raw.person_id = dat.person_id
                    AND COALESCE(raw.post_id, '') = COALESCE(dat.post_id, '')
                    AND COALESCE(raw.start_date, '1900-01-01') = COALESCE(dat.start_date, '1900-01-01')
                    AND COALESCE(raw.end_date, NOW()) = COALESCE(dat.end_date, NOW()))
              WHERE dat.organization_id IS NULL
                AND dat.person_id IS NULL
                AND dat.post_id IS NULL
        '''

        self.executeTransaction(find_new)

        cols = [
           'label',
           'role',
           'start_date',
           'end_date',
           'organization_id',
           'person_id',
           'post_id',
           'updated_at',
        ]

        insert_fields = ', '.join(c for c in cols)
        select_fields = ', '.join('raw.{}'.format(c) for c in cols)

        insert_new = '''
            INSERT INTO councilmatic_core_membership (
              {insert_fields}
            )
              SELECT {select_fields}
              FROM raw_membership AS raw
              JOIN new_membership AS new
                ON (raw.organization_id = new.organization_id
                    AND raw.person_id = new.person_id
                    AND COALESCE(raw.post_id, '') = COALESCE(new.post_id, '')
                    AND COALESCE(raw.start_date, '1900-01-01') = COALESCE(new.start_date, '1900-01-01')
                    AND COALESCE(raw.end_date, NOW()) = COALESCE(new.end_date, NOW()))
        '''.format(insert_fields=insert_fields,
                   select_fields=select_fields)

        self.executeTransaction(insert_new)

        new_count = self.connection.execute('select count(*) from new_membership').first().count

        self.log_message('Found {0} new membership'.format(new_count), style='SUCCESS')

    def add_new_bills(self):
        cols = [
            'ocd_id',
            'ocd_created_at',
            'ocd_updated_at',
            'description',
            'identifier',
            'bill_type',
            'classification',
            'source_url',
            'source_note',
            'subject',
            'from_organization_id',
            'full_text',
            'ocr_full_text',
            'abstract',
            'last_action_date',
            'legislative_session_id',
            'slug',
        ]

        self.add_entity_type('bill', cols=cols)

    def add_new_actions(self):
        self.executeTransaction('DROP TABLE IF EXISTS new_action')
        self.executeTransaction('''
            CREATE TABLE new_action (
                bill_id VARCHAR,
                "order" INTEGER
            )
        ''')

        find_new = '''
            INSERT INTO new_action
              SELECT
                raw.bill_id,
                raw."order"
              FROM raw_action AS raw
              LEFT JOIN councilmatic_core_action AS dat
                ON (raw.bill_id = dat.bill_id
                    AND raw."order" = dat."order")
              WHERE dat.bill_id IS NULL
                AND dat."order" IS NULL
        '''

        self.executeTransaction(find_new)

        cols = [
            'date',
            'classification',
            'description',
            'order',
            'bill_id',
            'organization_id',
        ]

        insert_fields = ', '.join('"{}"'.format(c) for c in cols)
        select_fields = ', '.join('raw."{}"'.format(c) for c in cols)

        insert_new = '''
            INSERT INTO councilmatic_core_action (
              {insert_fields},
              updated_at
            )
              SELECT {select_fields}, raw.updated_at
              FROM raw_action AS raw
              JOIN new_action AS new
                ON (raw.bill_id = new.bill_id
                    AND raw."order" = new."order")
        '''.format(insert_fields=insert_fields,
                   select_fields=select_fields)

        self.executeTransaction(insert_new)

        new_count = self.connection.execute('select count(*) from new_action').first().count

        self.log_message('Found {0} new action'.format(new_count), style='SUCCESS')

    def add_new_action_related_entity(self):
        self.executeTransaction('DROP TABLE IF EXISTS new_actionrelatedentity')
        self.executeTransaction('''
            CREATE TABLE new_actionrelatedentity (
                organization_ocd_id VARCHAR,
                person_ocd_id VARCHAR,
                action_id INTEGER
            )
        ''')

        cols = [
            'entity_type',
            'entity_name',
            'organization_ocd_id',
            'person_ocd_id',
            'action_id',
        ]

        find_new = '''
            INSERT INTO new_actionrelatedentity
              SELECT
                raw.organization_ocd_id,
                raw.person_ocd_id,
                raw.action_id
              FROM raw_actionrelatedentity AS raw
              LEFT JOIN councilmatic_core_actionrelatedentity AS dat
                ON (COALESCE(raw.organization_ocd_id, '') = COALESCE(dat.organization_ocd_id, '')
                    AND COALESCE(raw.person_ocd_id, '') = COALESCE(dat.person_ocd_id, '')
                    AND raw.action_id = dat.action_id)
              WHERE (dat.organization_ocd_id IS NULL
                     OR dat.person_ocd_id IS NULL)
                     AND dat.action_id IS NULL
        '''

        self.executeTransaction(find_new)

        insert_fields = ', '.join(c for c in cols)
        select_fields = ', '.join('raw.{}'.format(c) for c in cols)

        insert_new = '''
            INSERT INTO councilmatic_core_actionrelatedentity (
              {insert_fields}, updated_at
            )
              SELECT {select_fields}, updated_at
              FROM raw_actionrelatedentity AS raw
              JOIN new_actionrelatedentity AS new
                ON (COALESCE(raw.organization_ocd_id, '') = COALESCE(new.organization_ocd_id, '')
                    AND COALESCE(raw.person_ocd_id, '') = COALESCE(new.person_ocd_id, '')
                    AND raw.action_id = new.action_id)
        '''.format(insert_fields=insert_fields,
                   select_fields=select_fields)

        self.executeTransaction(insert_new)

        new_count = self.connection.execute('select count(*) from new_actionrelatedentity').first().count

        self.log_message('Found {0} new action related entities'.format(new_count), style='SUCCESS')

    def add_new_sponsorships(self):
        self.executeTransaction('DROP TABLE IF EXISTS new_sponsorship')
        self.executeTransaction('''
            CREATE TABLE new_sponsorship (
                classification VARCHAR,
                is_primary BOOLEAN,
                bill_id VARCHAR,
                person_id VARCHAR
            )
        ''')

        cols = [
            'classification',
            'is_primary',
            'bill_id',
            'person_id',
        ]

        find_new = '''
            INSERT INTO new_sponsorship
              SELECT
                raw.classification,
                raw.is_primary,
                raw.bill_id,
                raw.person_id
              FROM raw_sponsorship AS raw
              LEFT JOIN councilmatic_core_sponsorship AS dat
                ON (raw.classification = dat.classification
                    AND raw.is_primary = dat.is_primary
                    AND raw.bill_id = dat.bill_id
                    AND raw.person_id = dat.person_id)
              WHERE dat.classification IS NULL
                    AND dat.is_primary IS NULL
                    AND dat.bill_id IS NULL
                    AND dat.person_id IS NULL
        '''

        self.executeTransaction(find_new)

        insert_fields = ', '.join(c for c in cols)
        select_fields = ', '.join('raw.{}'.format(c) for c in cols)

        insert_new = '''
            INSERT INTO councilmatic_core_sponsorship (
              {insert_fields}, updated_at
            )
              SELECT {select_fields}, updated_at
              FROM raw_sponsorship AS raw
              JOIN new_sponsorship AS new
                ON (raw.classification = new.classification
                    AND raw.is_primary = new.is_primary
                    AND raw.bill_id = new.bill_id
                    AND raw.person_id = new.person_id)
        '''.format(insert_fields=insert_fields,
                   select_fields=select_fields)

        self.executeTransaction(insert_new)

        new_count = self.connection.execute('select count(*) from new_sponsorship').first().count

        self.log_message('Found {0} new sponsorships'.format(new_count), style='SUCCESS')

    def add_new_billdocuments(self):
        self.executeTransaction('DROP TABLE IF EXISTS new_billdocument')
        self.executeTransaction('''
            CREATE TABLE new_billdocument (
                bill_id VARCHAR,
                url VARCHAR,
                document_type VARCHAR
            )
        ''')

        cols = [
            'bill_id',
            'url',
            'document_type',
            'note',
        ]

        find_new = '''
            INSERT INTO new_billdocument
              SELECT
                raw.bill_id,
                raw.url,
                raw.document_type
              FROM raw_billdocument AS raw
              LEFT JOIN councilmatic_core_billdocument AS dat
                ON (raw.bill_id = dat.bill_id
                    AND raw.url = dat.url
                    AND raw.document_type = dat.document_type)
              WHERE dat.bill_id IS NULL
                    AND dat.url IS NULL
                    AND dat.document_type IS NULL
        '''

        self.executeTransaction(find_new)

        insert_fields = ', '.join(c for c in cols)
        select_fields = ', '.join('raw.{}'.format(c) for c in cols)

        insert_new = '''
            INSERT INTO councilmatic_core_billdocument (
              {insert_fields}, updated_at
            )
              SELECT {select_fields}, updated_at
              FROM raw_billdocument AS raw
              JOIN new_billdocument AS new
                ON (raw.bill_id = new.bill_id
                    AND raw.url = new.url
                    AND raw.document_type = new.document_type)
        '''.format(insert_fields=insert_fields,
                   select_fields=select_fields)

        self.executeTransaction(insert_new)

        new_count = self.connection.execute('select count(*) from new_billdocument').first().count

        self.log_message('Found {0} new bill documents'.format(new_count), style='SUCCESS')

    def add_new_events(self):
        cols = [
            'ocd_id',
            'ocd_created_at',
            'ocd_updated_at',
            'name',
            'description',
            'classification',
            'start_time',
            'end_time',
            'all_day',
            'status',
            'location_name',
            'location_url',
            'source_url',
            'source_note',
            'slug',
        ]

        self.add_entity_type('event', cols=cols)

    def add_new_eventparticipants(self):
        self.executeTransaction('DROP TABLE IF EXISTS new_eventparticipant')
        self.executeTransaction('''
            CREATE TABLE new_eventparticipant (
                event_id VARCHAR,
                entity_type VARCHAR,
                entity_name VARCHAR
            )
        ''')

        cols = [
            'event_id',
            'entity_type',
            'entity_name',
            'note',
        ]

        find_new = '''
            INSERT INTO new_eventparticipant
              SELECT
                raw.event_id,
                raw.entity_type,
                raw.entity_name
              FROM raw_eventparticipant AS raw
              LEFT JOIN councilmatic_core_eventparticipant AS dat
                ON (raw.event_id = dat.event_id
                    AND raw.entity_type = dat.entity_type
                    AND raw.entity_name = dat.entity_name)
              WHERE dat.event_id IS NULL
                    AND dat.entity_type IS NULL
                    AND dat.entity_name IS NULL
        '''

        self.executeTransaction(find_new)

        insert_fields = ', '.join(c for c in cols)
        select_fields = ', '.join('raw.{}'.format(c) for c in cols)

        insert_new = '''
            INSERT INTO councilmatic_core_eventparticipant (
              {insert_fields}, updated_at
            )
              SELECT {select_fields}, updated_at
              FROM raw_eventparticipant AS raw
              JOIN new_eventparticipant AS new
                ON (raw.event_id = new.event_id
                    AND raw.entity_type = new.entity_type
                    AND raw.entity_name = new.entity_name)
        '''.format(insert_fields=insert_fields,
                   select_fields=select_fields)

        self.executeTransaction(insert_new)

        new_count = self.connection.execute('select count(*) from new_eventparticipant').first().count

        self.log_message('Found {0} new event participants'.format(new_count), style='SUCCESS')

    def add_new_eventdocuments(self):
        self.executeTransaction('DROP TABLE IF EXISTS new_eventdocument')
        self.executeTransaction('''
            CREATE TABLE new_eventdocument (
                event_id VARCHAR,
                url VARCHAR
            )
        ''')

        cols = [
            'event_id',
            'url',
            'full_text',
            'note'
        ]

        find_new = '''
            INSERT INTO new_eventdocument
              SELECT
                raw.event_id,
                raw.url
              FROM raw_eventdocument AS raw
              LEFT JOIN councilmatic_core_eventdocument AS dat
                ON (raw.event_id = dat.event_id
                    AND raw.url = dat.url)
              WHERE dat.event_id IS NULL
                    AND dat.url IS NULL
        '''

        self.executeTransaction(find_new)

        insert_fields = ', '.join(c for c in cols)
        select_fields = ', '.join('raw.{}'.format(c) for c in cols)

        insert_new = '''
            INSERT INTO councilmatic_core_eventdocument (
              {insert_fields}
            )
              SELECT {select_fields}
              FROM raw_eventdocument AS raw
              JOIN new_eventdocument AS new
                ON (raw.event_id = new.event_id
                    AND raw.url = new.url)
        '''.format(insert_fields=insert_fields,
                   select_fields=select_fields)

        self.executeTransaction(insert_new)

        new_count = self.connection.execute('select count(*) from new_eventdocument').first().count

        self.log_message('Found {0} new event documents'.format(new_count), style='SUCCESS')

    def add_new_event_agenda_items(self):
        self.executeTransaction('DROP TABLE IF EXISTS new_eventagendaitem')
        self.executeTransaction('''
            CREATE TABLE new_eventagendaitem (
                event_id VARCHAR,
                "order" INTEGER
            )
        ''')

        cols = [
            'order',
            'description',
            'event_id',
            'bill_id',
            'note',
        ]

        find_new = '''
            INSERT INTO new_eventagendaitem
              SELECT
                raw.event_id,
                raw."order"
              FROM raw_eventagendaitem AS raw
              LEFT JOIN councilmatic_core_eventagendaitem AS dat
                ON (raw.event_id = dat.event_id
                    AND raw."order" = dat."order")
              WHERE dat.event_id IS NULL
                    AND dat."order" IS NULL
        '''

        self.executeTransaction(find_new)

        insert_fields = ', '.join('"{}"'.format(c) for c in cols)
        select_fields = ', '.join('raw."{}"'.format(c) for c in cols)

        insert_new = '''
            INSERT INTO councilmatic_core_eventagendaitem (
              {insert_fields}, updated_at
            )
              SELECT {select_fields}, updated_at
              FROM raw_eventagendaitem AS raw
              JOIN new_eventagendaitem AS new
                ON (raw.event_id = new.event_id
                    AND raw."order" = new."order")
        '''.format(insert_fields=insert_fields,
                   select_fields=select_fields)

        self.executeTransaction(insert_new)
        new_count = self.connection.execute('select count(*) from new_eventagendaitem').first().count

        self.log_message('Found {0} new event agenda items'.format(new_count), style='SUCCESS')


    def populate_council_district_shapes(self):

        self.log_message('Populating boundaries ...')
        # grab boundary listing
        for boundary in settings.BOUNDARY_SET:
            bndry_set_url = bndry_base_url + '/boundaries/' + boundary
            r = requests.get(bndry_set_url + '/?limit=0')
            page_json = json.loads(r.text)

            # loop through boundary listing
            for bndry_json in page_json['objects']:
                # grab boundary shape
                shape_url = bndry_base_url + bndry_json['url'] + 'shape'
                r = requests.get(shape_url)
                # update the right post(s) with the shape
                if 'ocd-division' in bndry_json['external_id']:
                    division_ocd_id = bndry_json['external_id']

                    Post.objects.filter(
                        division_ocd_id=division_ocd_id).update(shape=r.text)
                else:
                    # Represent API doesn't use OCD id as external_id,
                    # so we must work around that
                    division_ocd_id_fragment = ':' + bndry_json['external_id']
                    Post.objects.filter(
                        division_ocd_id__endswith=division_ocd_id_fragment).update(shape=r.text)

                print('.', end='')
                sys.stdout.flush()

    def executeTransaction(self, query, *args, **kwargs):
        trans = self.connection.begin()

        raise_exc = kwargs.get('raise_exc', True)

        try:
            self.connection.execute("SET local timezone to '{}'".format(settings.TIME_ZONE))
            if kwargs:
                self.connection.execute(query, **kwargs)
            else:
                self.connection.execute(query, *args)
            trans.commit()
        except sa.exc.ProgrammingError as e:
            # TODO: Make some kind of logger
            # logger.error(e, exc_info=True)
            trans.rollback()
            if raise_exc:
                raise e



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0004_auto_20151214_1413'),
    ]

    operations = [
        migrations.AddField(
            model_name='billdocument',
            name='document_type',
            field=models.CharField(max_length=255, choices=[(
                'A', 'Attachment'), ('V', 'Version')], default='A'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='document',
            name='full_text',
            field=models.TextField(blank=True),
        ),
        migrations.AlterField(
            model_name='document',
            name='url',
            field=models.TextField(blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-10-14 18:49
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0020_auto_20161013_1704'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='billdocument',
            name='document',
        ),
        migrations.RemoveField(
            model_name='eventdocument',
            name='document',
        ),
        migrations.RemoveField(
            model_name='eventdocument',
            name='updated_at',
        ),
        migrations.AddField(
            model_name='billdocument',
            name='full_text',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='billdocument',
            name='note',
            field=models.TextField(default=''),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='billdocument',
            name='url',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='eventdocument',
            name='full_text',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='eventdocument',
            name='note',
            field=models.TextField(default=''),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='eventdocument',
            name='url',
            field=models.TextField(blank=True),
        ),
        migrations.DeleteModel(
            name='Document',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-10-12 19:19
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0013_auto_20161012_1414'),
    ]

    operations = [
        migrations.AlterField(
            model_name='post',
            name='shape',
            field=models.TextField(blank=True, null=True),
        ),
    ]



from .models import Person, Bill, Organization, Event, Post
from datetime import date, timedelta, datetime
import itertools
from operator import attrgetter
import urllib
from django.contrib.syndication.views import Feed
from django.utils.feedgenerator import Rss201rev2Feed
from django.core.urlresolvers import reverse, reverse_lazy
from django.conf import settings

import pytz
import re
import json

from haystack.query import SearchQuerySet

class CouncilmaticFacetedSearchFeed(Feed):
    title_template = 'feeds/search_item_title.html'
    description_template = 'feeds/search_item_description.html'
    bill_model = Bill

    all_results = None
    sqs = SearchQuerySet().facet('bill_type')\
                          .facet('sponsorships', sort='index')\
                          .facet('controlling_body')\
                          .facet('inferred_status')
    query = None

    def url_with_querystring(self, path, **kwargs):
        return path + '?' + urllib.parse.urlencode(kwargs)

    def get_object(self, request):
        self.queryDict = request.GET
        all_results    = SearchQuerySet().all()

        if 'selected_facets' in request.GET:
            facets = request.GET.getlist('selected_facets')

        if 'q' in request.GET:
            self.query = request.GET['q']
            results = all_results.filter(content=self.query)

            if facets:
                for facet in facets:
                    (facet_name, facet_value) = facet.split(':')
                    facet_name = facet_name.rsplit('_exact')[0]
                    results = results.narrow('%s:%s' % (facet_name, facet_value))
        elif facets:
            for facet in facets:
                (facet_name, facet_value) = facet.split(':')
                facet_name = facet_name.rsplit('_exact')[0]
                results = all_results.narrow('%s:%s' % (facet_name, facet_value))

        return results.order_by('-last_action_date')

    def title(self, obj):
        if self.query:
            title = settings.SITE_META['site_name'] + ": Search for '" + self.query.capitalize() + "'"
            # XXX: create a nice title based on all search parameters
        else:
            title = settings.SITE_META['site_name'] + ": Filtered Search"

        return title

    def link(self, obj):
        # return the main non-RSS search URL somehow
        # XXX maybe "quargs" - evz
        # return reverse('councilmatic_search', args=(searchqueryset=self.sqs,))
        url = self.url_with_querystring(reverse('{}:councilmatic_search_feed'.format(settings.APP_NAME)), q=self.query)
        return url

    def item_link(self, bill):
        return reverse('bill_detail', args=(bill.slug,))

    def item_pubdate(self, bill):
        return bill.last_action_date

    def description(self, obj):
        return "Bills returned from search"

    def items(self, searchresults):
        l_items = list(searchresults)[:20]
        # turn these into bills. XXX: should override in subclasses, e.g. NYCCouncilmaticFacetedSearchFeed,
        # to access methods like inferred_status()
        pks =[i.pk for i in l_items]
        bills = self.bill_model.objects.filter(pk__in=pks).order_by('-last_action_date')
        return list(bills)


class PersonDetailFeed(Feed):
    """The PersonDetailFeed provides an RSS feed for a given committee member,
    returning the most recent 20 bills for which they are the primary sponsor;
    and for each bill, the list of sponsores and the action history.
    """

    title_template = 'feeds/person_detail_item_title.html'
    description_template = 'feeds/person_detail_item_description.html'
    feed_type = Rss201rev2Feed
    NUM_RECENT_BILLS = 20

    def get_object(self, request, slug):
        o = Person.objects.get(slug=slug)
        return o

    def title(self, obj):
        return settings.SITE_META['site_name'] + ': ' + settings.CITY_VOCAB['COUNCIL_MEMBER'] + " %s: Recently Sponsored Bills" % obj.name

    def link(self, obj):
        # return the Councilmatic URL for the person
        # XXX maybe put this in models.py:Person.get_absolute_url() instead (https://docs.djangoproject.com/en/1.9/ref/models/instances/ , https://docs.djangoproject.com/en/1.9/ref/contrib/syndication/)
        return reverse('person', args=(obj.slug,))

    # Return sponsored legislation a la https://nyc.councilmatic.org/person/margaret-s-chin/
    def item_link(self, bill):
        # return the Councilmatic URL for the bill
        return reverse('bill_detail', args=(bill.slug,))

    def item_pubdate(self, bill):
        return bill.last_action_date

    def description(self, obj):
        return "Recent sponsored bills from " + obj.name + "."

    def items(self, person):
        sponsored_bills = [s.bill for s in person.primary_sponsorships.order_by('-_bill__last_action_date')[:10]]
        recent_sponsored_bills = sponsored_bills[:self.NUM_RECENT_BILLS]
        return recent_sponsored_bills


class CommitteeDetailEventsFeed(Feed):
    """The CommitteeDetailEventsFeed provides an RSS feed for a given committee,
    returning the most recent 20 events.
    """

    title_template = 'feeds/committee_events_item_title.html'
    description_template = 'feeds/committee_events_item_description.html'
    feed_type = Rss201rev2Feed
    NUM_RECENT_COMMITTEE_EVENTS = 20

    def get_object(self, request, slug):
        o = Organization.objects.get(slug=slug)
        return o

    def title(self, obj):
        return settings.SITE_META['site_name'] + ": " + obj.name + ": Recent Events"

    def link(self, obj):
        # return the Councilmatic URL for the committee
        return reverse('committee_detail', args=(obj.slug,))

    def item_link(self, event):
        # return the Councilmatic URL for the event
        return reverse('event_detail', args=(event.slug,))

    def item_pubdate(self, event):
        return event.start_time

    def description(self, obj):
        return "Events for committee %s" % obj.name

    def items(self, obj):
        events =  obj.recent_events.all()[:self.NUM_RECENT_COMMITTEE_EVENTS]
        levents =  list(events)
        return levents


class CommitteeDetailActionFeed(Feed):
    """The CommitteeDetailActionFeed provides an RSS feed for a given committee,
    returning the most recent 20 actions on legislation.
    """

    # instead of defining item_title() or item_description(), use templates
    title_template = 'feeds/committee_actions_item_title.html'
    description_template = 'feeds/committee_actions_item_description.html'
    feed_type = Rss201rev2Feed
    NUM_RECENT_COMMITTEE_ACTIONS = 20

    def get_object(self, request, slug):
        o = Organization.objects.get(slug=slug)
        return o

    def title(self, obj):
        return settings.SITE_META['site_name'] + ": " + obj.name + ": Recent Actions"

    def link(self, obj):
        # return the Councilmatic URL for the committee
        return reverse('committee_detail', args=(obj.slug,))

    def item_link(self, action):
        # return the Councilmatic URL for the bill
        return reverse('bill_detail', args=(action.bill.slug,))

    def item_pubdate(self, action):
        return action.date

    def description(self, obj):
        return "Actions for committee %s" % obj.name

    def items(self, obj):
        actions = obj.recent_activity[:self.NUM_RECENT_COMMITTEE_ACTIONS]
        actions_list =  list(actions)
        return actions_list

class BillDetailActionFeed(Feed):
    """
    Return the last 20 actions for a given bill.
    """

    # instead of defining item_title() or item_description(), use templates
    title_template = 'feeds/bill_actions_item_title.html'
    description_template = 'feeds/bill_actions_item_description.html'
    feed_type = Rss201rev2Feed
    NUM_RECENT_BILL_ACTIONS = 20

    def get_object(self, request, slug):
        o = Bill.objects.get(slug=slug)
        return o

    def title(self, obj):
        return settings.SITE_META['site_name'] + ": " + obj.friendly_name + ": Recent Actions"

    def link(self, obj):
        # return the Councilmatic URL for the committee
        return reverse('bill_detail', args=(obj.slug,))

    def item_link(self, action):
        # Bill actions don't have their own pages, so just link to the Bill page (?)
        return reverse('bill_detail', args=(action.bill.slug,))

    def item_pubdate(self, action):
        return action.date

    def description(self, obj):
        return "Actions for bill %s" % obj.friendly_name

    def items(self, obj):
        actions = obj.ordered_actions[:self.NUM_RECENT_BILL_ACTIONS]
        actions_list =  list(actions)
        return actions_list


class EventsFeed(Feed):
    """
    Return the last 20 announced events as per, e.g., https://nyc.councilmatic.org/events/
    """
    title_template = 'feeds/events_item_title.html'
    description_template = 'feeds/events_item_description.html'
    feed_type = Rss201rev2Feed
    NUM_RECENT_EVENTS = 20


    #def get_object(self, request):
    #    o = Event.objects.all()
    #    return o

    title = settings.CITY_COUNCIL_NAME + " " + "Recent Events"
    link = reverse_lazy('events')
    description = "Recently announced events."

    def item_link(self, event):
        # return the Councilmatic URL for the event
        return reverse('event_detail', args=(event.slug,))

    def item_pubdate(self, event):
        return event.start_time

    def description(self, obj):
        return "Events"

    def items(self, obj):
        events =  Event.objects.all()[:self.NUM_RECENT_EVENTS]
        levents =  list(events)
        return levents




import re
import json
import itertools
from operator import attrgetter
import urllib
from datetime import date, timedelta, datetime

from django.shortcuts import render, redirect
from django.conf import settings
from django.views.generic import TemplateView, ListView, DetailView
from django.views.decorators.clickjacking import xframe_options_exempt
from django.db.models import Max, Min
from django.core.cache import cache
from django.utils.text import slugify
from django.utils.decorators import method_decorator
from django.utils import timezone


from django.template import *

from haystack.forms import FacetedSearchForm
from haystack.views import FacetedSearchView
from haystack.query import SearchQuerySet

import pytz

from .models import Person, Bill, Organization, Event, Post

from django.views.decorators.cache import never_cache
from django.contrib.auth.decorators import login_required
from django.utils.decorators import method_decorator

if (settings.USING_NOTIFICATIONS):
    from notifications.models import BillSearchSubscription, PersonSubscription

app_timezone = pytz.timezone(settings.TIME_ZONE)

# XXX mcc: genuinely not sure if this is helping or not
class NeverCacheMixin(object):
    @method_decorator(never_cache)
    def dispatch(self, *args, **kwargs):
        return super(NeverCacheMixin, self).dispatch(*args, **kwargs)

class CouncilmaticFacetedSearchView(FacetedSearchView, NeverCacheMixin):

    def extra_context(self):

        extra = super(FacetedSearchView, self).extra_context()
        extra['request'] = self.request
        extra['facets'] = self.results.facet_counts()

        q_filters = ''
        url_params = [(p, val) for (p, val) in self.request.GET.items(
        ) if p != 'page' and p != 'selected_facets' and p != 'amp' and p != '_']
        selected_facet_vals = self.request.GET.getlist('selected_facets')
        search_term = self.request.GET.get('q')
        for facet_val in selected_facet_vals:
            url_params.append(('selected_facets', facet_val))
        if url_params:
            q_filters = urllib.parse.urlencode(url_params)

        extra['q_filters'] = q_filters

        selected_facets = {}
        for val in self.request.GET.getlist("selected_facets"):
            if val:
                [k, v] = val.split('_exact:', 1)
                try:
                    selected_facets[k].append(v)
                except KeyError:
                    selected_facets[k] = [v]

        extra['selected_facets'] = selected_facets

        extra['current_council_members'] = {
            p.current_member.person.name: p.label for p in Post.objects.all() if p.current_member
        }

        if (settings.USING_NOTIFICATIONS):
            extra['user_subscribed'] = False
            if self.request.user.is_authenticated():
                user = self.request.user
                extra['user'] = user

                search_params = {
                    'term': search_term,
                    'facets': selected_facets
                }

                try:
                    bss = user.billsearchsubscriptions.get(user=user,
                                                           search_params__exact=search_params)
                    extra['user_subscribed'] = True
                except BillSearchSubscription.DoesNotExist:
                    extra['user_subscribed'] = False

        return extra


class CouncilmaticSearchForm(FacetedSearchForm):

    def __init__(self, *args, **kwargs):
        self.load_all = True

        super(CouncilmaticSearchForm, self).__init__(*args, **kwargs)

    def no_query_found(self):

        # return self.searchqueryset.order_by('-last_action_date').all()
        return self.searchqueryset.all()

# This is used by a context processor in settings.py to render these variables
# into the context of every page.

def city_context(request):
    relevant_settings = [
        'SITE_META',
        'FOOTER_CREDITS',
        'CITY_COUNCIL_NAME',
        'CITY_NAME',
        'CITY_NAME_SHORT',
        'CITY_VOCAB',
        'SEARCH_PLACEHOLDER_TEXT',
        'LEGISLATION_TYPE_DESCRIPTIONS',
        'LEGISTAR_URL',
        'DISQUS_SHORTNAME',
        'IMAGES',
        'MAP_CONFIG',
        'ANALYTICS_TRACKING_CODE',
        'ABOUT_BLURBS',
        'USING_NOTIFICATIONS'
    ]

    city_context = {s: getattr(settings, s, None) for s in relevant_settings}

    return city_context


class IndexView(TemplateView):

    template_name = 'councilmatic_core/index.html'
    bill_model = Bill
    event_model = Event

    def get_context_data(self, **kwargs):
        context = super(IndexView, self).get_context_data(**kwargs)

        recently_passed = []
        # go back in time at 10-day intervals til you find 3 passed bills
        for i in range(0, -100, -10):

            today = timezone.now()

            begin = today + timedelta(days=i)
            end = today + timedelta(days=i - 10)

            leg_in_range = self.bill_model.objects\
                .exclude(last_action_date=None)\
                .filter(last_action_date__lte=begin)\
                .filter(last_action_date__gt=end)\
                .order_by('-last_action_date')
            passed_in_range = [l for l in leg_in_range
                               if l.inferred_status == 'Passed']

            recently_passed.extend(passed_in_range)
            if len(recently_passed) >= 3:
                recently_passed = recently_passed[:3]
                break

        upcoming_meetings = list(
            self.event_model.upcoming_committee_meetings())

        context.update({
            'recently_passed': recently_passed,
            'next_council_meeting': self.event_model.next_city_council_meeting(),
            'upcoming_committee_meetings': upcoming_meetings,
        })

        context.update(self.extra_context())

        return context

    def extra_context(self):
        """
        Override this in custom subclass to add more context variables if needed.
        """
        return {}


class AboutView(TemplateView):
    template_name = 'councilmatic_core/about.html'


class CouncilMembersView(ListView):
    template_name = 'councilmatic_core/council_members.html'
    context_object_name = 'posts'

    def get_queryset(self):
        if hasattr(settings, 'OCD_CITY_COUNCIL_ID'):
            get_kwarg = {'ocd_id': settings.OCD_CITY_COUNCIL_ID}
        else:
            get_kwarg = {'name': settings.OCD_CITY_COUNCIL_NAME}

        return Organization.objects.get(**get_kwarg).posts.all()

    def get_context_data(self, *args, **kwargs):
        context = super(CouncilMembersView, self).get_context_data(**kwargs)
        context['seo'] = self.get_seo_blob()

        context['map_geojson'] = None

        if settings.MAP_CONFIG:
            map_geojson = {
                'type': 'FeatureCollection',
                'features': []
            }

            for post in self.object_list:
                if post.shape:

                    council_member = "Vacant"
                    detail_link = ""
                    if post.current_member:
                        council_member = post.current_member.person.name
                        detail_link = post.current_member.person.slug

                    feature = {
                        'type': 'Feature',
                        'geometry': json.loads(post.shape),
                        'properties': {
                            'district': post.label,
                            'council_member': council_member,
                            'detail_link': '/person/' + detail_link,
                            'select_id': 'polygon-{}'.format(slugify(post.label)),
                        }
                    }

                    map_geojson['features'].append(feature)

            context['map_geojson'] = json.dumps(map_geojson)

        return context

    def get_seo_blob(self):
        seo = {}
        seo.update(settings.SITE_META)
        return seo


class BillDetailView(DetailView):
    model = Bill
    template_name = 'councilmatic_core/legislation.html'
    context_object_name = 'legislation'

    def get_context_data(self, **kwargs):
        context = super(BillDetailView, self).get_context_data(**kwargs)

        context['actions'] = self.get_object().actions.all().order_by('-order')
        bill = context['legislation']

        seo = {}
        seo.update(settings.SITE_META)
        seo['site_desc'] = bill.listing_description
        seo['title'] = '%s - %s' % (bill.friendly_name,
                                    settings.SITE_META['site_name'])
        context['seo'] = seo

        context['user_subscribed'] = False
        if self.request.user.is_authenticated():
            user = self.request.user
            context['user'] = user
            # check if person of interest is subscribed to by user

            if settings.USING_NOTIFICATIONS:
                for bas in user.billactionsubscriptions.all():

                    if bill == bas.bill:
                        context['user_subscribed'] = True
                        break

        return context


@method_decorator(xframe_options_exempt, name='dispatch')
class BillWidgetView(BillDetailView):
    template_name = 'councilmatic_core/widgets/legislation.html'


class CommitteesView(ListView):
    template_name = 'councilmatic_core/committees.html'
    context_object_name = 'committees'

    def get_queryset(self):
        return Organization.committees


class CommitteeDetailView(DetailView):
    model = Organization
    template_name = 'councilmatic_core/committee.html'
    context_object_name = 'committee'

    def get_context_data(self, **kwargs):
        context = super(CommitteeDetailView, self).get_context_data(**kwargs)

        committee = context['committee']
        context['memberships'] = committee.memberships.all()

        description = None

        if getattr(settings, 'COMMITTEE_DESCRIPTIONS', None):
            description = settings.COMMITTEE_DESCRIPTIONS.get(committee.slug)
            context['committee_description'] = description

        seo = {}
        seo.update(settings.SITE_META)

        if description:
            seo['site_desc'] = description
        else:
            seo['site_desc'] = "See what %s's %s has been up to!" % (
                settings.CITY_COUNCIL_NAME, committee.name)

        seo['title'] = '%s - %s' % (committee.name,
                                    settings.SITE_META['site_name'])
        context['seo'] = seo

        context['user_subscribed_actions'] = False
        context['user_subscribed_events'] = False
        if self.request.user.is_authenticated():
            user = self.request.user
            context['user'] = user
            # check if person of interest is subscribed to by user

            if settings.USING_NOTIFICATIONS:
                for cas in user.committeeactionsubscriptions.all():
                    if committee == cas.committee:
                        context['user_subscribed_actions'] = True
                for ces in user.committeeeventsubscriptions.all():
                    if committee == ces.committee:
                        context['user_subscribed_events'] = True

        return context

@method_decorator(xframe_options_exempt, name='dispatch')
class CommitteeWidgetView(CommitteeDetailView):
    template_name = 'councilmatic_core/widgets/committee.html'


class PersonDetailView(DetailView):
    model = Person
    template_name = 'councilmatic_core/person.html'
    context_object_name = 'person'

    def get_context_data(self, **kwargs):
        context = super(PersonDetailView, self).get_context_data(**kwargs)

        person = context['person']
        context['sponsored_legislation'] = [
            s.bill for s in person.primary_sponsorships.order_by('-_bill__last_action_date')[:10]]

        title = ''
        if person.current_council_seat:
            title = '%s %s' % (person.current_council_seat,
                               settings.CITY_VOCAB['COUNCIL_MEMBER'])
        elif person.latest_council_seat:
            title = 'Former %s, %s' % (
                settings.CITY_VOCAB['COUNCIL_MEMBER'], person.latest_council_seat)
        elif getattr(settings, 'EXTRA_TITLES', None) and person.slug in settings.EXTRA_TITLES:
            title = settings.EXTRA_TITLES[person.slug]
        context['title'] = title

        seo = {}
        seo.update(settings.SITE_META)
        if person.current_council_seat:
            short_name = re.sub(r',.*', '', person.name)
            seo['site_desc'] = '%s - %s representative in %s. See what %s has been up to!' % (
                person.name, person.current_council_seat, settings.CITY_COUNCIL_NAME, short_name)
        else:
            seo['site_desc'] = 'Details on %s, %s' % (
                person.name, settings.CITY_COUNCIL_NAME)
        seo['title'] = '%s - %s' % (person.name,
                                    settings.SITE_META['site_name'])
        seo['image'] = person.headshot_url
        context['seo'] = seo

        context['map_geojson'] = None

        if settings.MAP_CONFIG and person.latest_council_membership and person.latest_council_membership.post.shape:
            map_geojson = {
                'type': 'FeatureCollection',
                'features': []
            }

            feature = {
                'type': 'Feature',
                'geometry': json.loads(person.latest_council_membership.post.shape),
                'properties': {
                    'district': person.latest_council_membership.post.label,
                }
            }

            map_geojson['features'].append(feature)

            context['map_geojson'] = json.dumps(map_geojson)

        context['user_subscribed'] = False

        if (settings.USING_NOTIFICATIONS):
            subscriptions = PersonSubscription.objects.all()

            if self.request.user.is_authenticated():
                user = self.request.user
                context['user'] = user
                # check if person of interest is subscribed to by user

                for ps in user.personsubscriptions.all():
                    if person == ps.person:
                        context['user_subscribed'] = True
                        break

        return context

@method_decorator(xframe_options_exempt, name='dispatch')
class PersonWidgetView(PersonDetailView):
    template_name = 'councilmatic_core/widgets/person.html'


class EventsView(ListView):
    template_name = 'councilmatic_core/events.html'

    def get_queryset(self):
        # Realize this is stupid. The reason this exists is so that
        # we can have this be a ListView in the inherited subclasses
        # if needed
        return []

    def get_context_data(self, **kwargs):
        context = super(EventsView, self).get_context_data(**kwargs)

        aggregates = Event.objects.aggregate(
            Min('start_time'), Max('start_time'))
        min_year, max_year = aggregates[
            'start_time__min'].year, aggregates['start_time__max'].year
        context['year_range'] = list(reversed(range(min_year, max_year + 1)))

        context['month_options'] = []
        for index in range(1, 13):
            month_name = datetime(datetime.now(app_timezone).date().year, index, 1).strftime('%B')
            context['month_options'].append([month_name, index])

        context['show_upcoming'] = True
        context['this_month'] = datetime.now(app_timezone).date().month
        context['this_year'] = datetime.now(app_timezone).date().year
        events_key = 'upcoming_events'

        upcoming_dates = Event.objects.filter(start_time__gt=datetime.now(app_timezone).date())

        current_year = self.request.GET.get('year')
        current_month = self.request.GET.get('month')

        # if there are no upcoming events, show the most recent month
        # that has events populated
        if not upcoming_dates and not current_year and not current_month:

            most_recent_past_starttime = Event.objects.order_by(
                '-start_time').first().start_time
            current_year = most_recent_past_starttime.year
            current_month = most_recent_past_starttime.month

        if current_year and current_month:
            events_key = 'month_events'
            upcoming_dates = Event.objects\
                                  .filter(start_time__year=int(current_year))\
                                  .filter(start_time__month=int(current_month))

            context['show_upcoming'] = False
            context['this_month'] = int(current_month)
            context['this_year'] = int(current_year)
            context['this_start_date']= datetime(year=int(current_year), month=int(current_month), day=1)

        context['user_subscribed'] = False
        if self.request.user.is_authenticated():
            user = self.request.user
            context['user'] = user

            if settings.USING_NOTIFICATIONS:
                if (len(user.eventssubscriptions.all()) > 0):
                    context['user_subscribed'] = True

        upcoming_dates = upcoming_dates.order_by('start_time')

        day_grouper = lambda x: (
            x.start_time.year, x.start_time.month, x.start_time.day)
        context[events_key] = []

        for event_date, events in itertools.groupby(upcoming_dates, key=day_grouper):
            events = sorted(events, key=attrgetter('start_time'))
            context[events_key].append([date(*event_date), events])

        return context


class EventDetailView(DetailView):
    template_name = 'councilmatic_core/event.html'
    model = Event
    context_object_name = 'event'

    def get_context_data(self, **kwargs):
        context = super(EventDetailView, self).get_context_data(**kwargs)
        event = context['event']

        participants = [p.entity_name for p in event.participants.all()]
        context['participants'] = Organization.objects.filter(
            name__in=participants)

        seo = {}
        seo.update(settings.SITE_META)
        seo['site_desc'] = 'Public city council event on %s/%s/%s - view event participants & agenda items' % (
            event.start_time.month, event.start_time.day, event.start_time.year)
        seo['title'] = '%s Event - %s' % (event.name,
                                          settings.SITE_META['site_name'])
        context['seo'] = seo

        return context


def flush(request, flush_key):

    try:
        if flush_key == settings.FLUSH_KEY:
            cache.clear()
    except AttributeError:
        print("\n\n** NOTE: to use flush-cache, set FLUSH_KEY in settings_local.py **\n\n")

    return redirect('index')

@xframe_options_exempt
def pdfviewer(request):
    return render(request, 'councilmatic_core/pdfviewer.html')
